import io
import logging
import re
import time
import warnings
from typing import List

import numpy as np
import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis.strategies import floats, tuples

import qcodes.instrument.sims as sims
from qcodes.instrument.base import Instrument
from qcodes.instrument.ip_to_visa import AMI430_VISA
from qcodes.instrument_drivers.american_magnetics.AMI430 import (
    AMI430,
    AMI430_3D,
    AMI430Warning,
)
from qcodes.math_utils.field_vector import FieldVector
from qcodes.utils.types import (
    numpy_concrete_floats,
    numpy_concrete_ints,
    numpy_non_concrete_floats_instantiable,
    numpy_non_concrete_ints_instantiable,
)

_time_resolution = time.get_clock_info('time').resolution

# If any of the field limit functions are satisfied we are in the safe zone.
# We can have higher field along the z-axis if x and y are zero.
field_limit = [
    lambda x, y, z: x == 0 and y == 0 and z < 3,
    lambda x, y, z: np.linalg.norm([x, y, z]) < 2
]

# path to the .yaml file containing the simulated instrument
visalib = sims.__file__.replace('__init__.py', 'AMI430.yaml@sim')


@pytest.fixture(scope='function')
def magnet_axes_instances():
    """
    Start three mock instruments representing current drivers for the x, y,
    and z directions.
    """
    mag_x = AMI430_VISA('x', address='GPIB::1::INSTR', visalib=visalib,
                        terminator='\n', port=1)
    mag_y = AMI430_VISA('y', address='GPIB::2::INSTR', visalib=visalib,
                        terminator='\n', port=1)
    mag_z = AMI430_VISA('z', address='GPIB::3::INSTR', visalib=visalib,
                        terminator='\n', port=1)

    yield mag_x, mag_y, mag_z

    mag_x.close()
    mag_y.close()
    mag_z.close()


@pytest.fixture(scope='function')
def current_driver(magnet_axes_instances):
    """
    Instantiate AMI430_3D instrument with the three mock instruments
    representing current drivers for the x, y, and z directions.
    """
    mag_x, mag_y, mag_z = magnet_axes_instances

    driver = AMI430_3D("AMI430-3D", mag_x, mag_y, mag_z, field_limit)

    yield driver

    driver.close()


@pytest.fixture(scope='function')
def ami430():
    mag = AMI430_VISA('ami430', address='GPIB::1::INSTR', visalib=visalib,
                      terminator='\n', port=1)
    yield mag
    mag.close()


# here the original test has a homemade log system that we don't want to
# reproduce / write tests for. Instead, we use normal logging from our
# instrument.visa module
iostream = io.StringIO()
logger = logging.getLogger('qcodes.instrument.visa')
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(created)s - %(message)s')
lh = logging.StreamHandler(iostream)
logger.addHandler(lh)
lh.setLevel(logging.DEBUG)
lh.setFormatter(formatter)


random_coordinates = {
    "cartesian": tuples(
        floats(min_value=0, max_value=1),  # x
        floats(min_value=0, max_value=1),  # y
        floats(min_value=0, max_value=1)  # z
    ),
    "spherical": tuples(
        floats(min_value=0, max_value=1),  # r
        floats(min_value=0, max_value=180),  # theta
        floats(min_value=0, max_value=180)  # phi
    ),
    "cylindrical": tuples(
        floats(min_value=0, max_value=1),  # rho
        floats(min_value=0, max_value=180),  # phi
        floats(min_value=0, max_value=1)  # z
    )
}


def test_instantiation_from_names(magnet_axes_instances, request):
    """
    Instantiate AMI430_3D instrument from the three mock instruments
    representing current drivers for the x, y, and z directions by their
    names as opposed from their instances.
    """
    mag_x, mag_y, mag_z = magnet_axes_instances
    request.addfinalizer(AMI430_3D.close_all)

    driver = AMI430_3D("AMI430-3D", mag_x.name, mag_y.name, mag_z.name,
                       field_limit)

    assert driver._instrument_x is mag_x
    assert driver._instrument_y is mag_y
    assert driver._instrument_z is mag_z


def test_instantiation_from_name_of_nonexistent_ami_instrument(
        magnet_axes_instances, request
):
    mag_x, mag_y, mag_z = magnet_axes_instances
    request.addfinalizer(AMI430_3D.close_all)

    non_existent_instrument = mag_y.name + "foo"

    with pytest.raises(
            KeyError,
            match=f"with name {non_existent_instrument} does not exist"
    ):
        AMI430_3D(
            "AMI430-3D",
            mag_x.name, non_existent_instrument, mag_z.name,
            field_limit
        )


def test_instantiation_from_name_of_existing_non_ami_instrument(
        magnet_axes_instances, request
):
    mag_x, mag_y, mag_z = magnet_axes_instances
    request.addfinalizer(AMI430_3D.close_all)

    non_ami_existing_instrument = Instrument("foo")

    with pytest.raises(
            TypeError,
            match=re.escape(
                f"Instrument {non_ami_existing_instrument.name} is "
                f"{type(non_ami_existing_instrument)} but {AMI430} "
                f"was requested"
            )
    ):
        AMI430_3D(
            "AMI430-3D",
            mag_x.name, non_ami_existing_instrument.name, mag_z.name,
            field_limit
        )


def test_instantiation_from_badly_typed_argument(
        magnet_axes_instances, request
):
    mag_x, mag_y, mag_z = magnet_axes_instances
    request.addfinalizer(AMI430_3D.close_all)

    badly_typed_instrument_z_argument = 123

    with pytest.raises(
            ValueError, match="instrument_z argument is neither of those"
    ):
        AMI430_3D(
            "AMI430-3D",
            mag_x.name, mag_y, badly_typed_instrument_z_argument,
            field_limit
        )


@given(set_target=random_coordinates["cartesian"])
@settings(max_examples=10, suppress_health_check=(HealthCheck.function_scoped_fixture,))
def test_cartesian_sanity(current_driver, set_target):
    """
    A sanity check to see if the driver remember vectors in any random
    configuration in cartesian coordinates
    """
    current_driver.cartesian(set_target)
    get_target = current_driver.cartesian()

    assert np.allclose(set_target, get_target)

    # test that we can get the individual coordinates
    x = current_driver.x()
    y = current_driver.y()
    z = current_driver.z()
    assert np.allclose(set_target, [x, y, z])


@given(set_target=random_coordinates["spherical"])
@settings(max_examples=10, suppress_health_check=(HealthCheck.function_scoped_fixture,))
def test_spherical_sanity(current_driver, set_target):
    """
    A sanity check to see if the driver remember vectors in any random
    configuration in spherical coordinates
    """
    current_driver.spherical(set_target)
    get_target = current_driver.spherical()

    assert np.allclose(set_target, get_target)

    # test that we can get the individual coordinates
    r = current_driver.field()
    theta = current_driver.theta()
    phi = current_driver.phi()
    assert np.allclose(set_target, [r, theta, phi])


@given(set_target=random_coordinates["cylindrical"])
@settings(max_examples=10, suppress_health_check=(HealthCheck.function_scoped_fixture,))
def test_cylindrical_sanity(current_driver, set_target):
    """
    A sanity check to see if the driver remember vectors in any random
    configuration in cylindrical coordinates
    """
    current_driver.cylindrical(set_target)
    get_target = current_driver.cylindrical()

    assert np.allclose(set_target, get_target)

    # test that we can get the individual coordinates
    rho = current_driver.rho()
    z = current_driver.z()
    phi = current_driver.phi()
    assert np.allclose(set_target, [rho, phi, z])


@given(set_target=random_coordinates["cartesian"])
@settings(max_examples=10, suppress_health_check=(HealthCheck.function_scoped_fixture,))
def test_cartesian_setpoints(current_driver, set_target):
    """
    Check that the individual x, y, z instruments are getting the set
    points as intended. This test is very similar to the sanity test, but
    adds in the FieldVector as well.
    """
    current_driver.cartesian(set_target)

    x = current_driver.x()
    y = current_driver.y()
    z = current_driver.z()

    get_target = dict(zip(('x', 'y', 'z'), (x, y, z)))

    set_vector = FieldVector(*set_target)
    get_vector = FieldVector(**get_target)
    assert set_vector.is_equal(get_vector)


@given(set_target=random_coordinates["spherical"])
@settings(max_examples=10, suppress_health_check=(HealthCheck.function_scoped_fixture,))
def test_spherical_setpoints(current_driver, set_target):
    """
    Check that the individual x, y, z instruments are getting the set
    points as intended. This test is very similar to the sanity test, but
    adds in the FieldVector as well.
    """
    current_driver.spherical(set_target)

    r = current_driver.field()
    theta = current_driver.theta()
    phi = current_driver.phi()

    get_target = dict(zip(('r', 'theta', 'phi'), (r, theta, phi)))
    set_target = dict(zip(('r', 'theta', 'phi'), set_target))

    set_vector = FieldVector(**set_target)
    get_vector = FieldVector(**get_target)
    assert set_vector.is_equal(get_vector)


@given(set_target=random_coordinates["cylindrical"])
@settings(max_examples=10, deadline=500,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
def test_cylindrical_setpoints(current_driver, set_target):
    """
    Check that the individual x, y, z instruments are getting the set
    points as intended. This test is very similar to the sanity test, but
    adds in the FieldVector as well.
    """
    current_driver.cylindrical(set_target)

    rho = current_driver.rho()
    z = current_driver.z()
    phi = current_driver.phi()

    get_target = dict(zip(('rho', 'phi', 'z'), (rho, phi, z)))
    set_target = dict(zip(('rho', 'phi', 'z'), set_target))

    set_vector = FieldVector(**set_target)
    get_vector = FieldVector(**get_target)
    assert set_vector.is_equal(get_vector)


@given(set_target=random_coordinates["cartesian"])
@settings(max_examples=10, deadline=500,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
def test_measured(current_driver, set_target):
    """
    Simply call the measurement methods and verify that no exceptions
    are raised.
    """
    current_driver.cartesian(set_target)

    cartesian = current_driver.cartesian_measured()

    cartesian_x = current_driver.x_measured()
    cartesian_y = current_driver.y_measured()
    cartesian_z = current_driver.z_measured()

    assert np.allclose(cartesian, [cartesian_x, cartesian_y, cartesian_z])
    assert FieldVector(*set_target).is_equal(FieldVector(x=cartesian_x,
                                                         y=cartesian_y,
                                                         z=cartesian_z))

    spherical = current_driver.spherical_measured()
    spherical_field = current_driver.field_measured()
    spherical_theta = current_driver.theta_measured()
    spherical_phi = current_driver.phi_measured()

    assert FieldVector(*set_target).is_equal(FieldVector(
        r=spherical_field,
        theta=spherical_theta,
        phi=spherical_phi)
    )
    assert np.allclose(spherical,
                       [spherical_field, spherical_theta, spherical_phi])

    cylindrical = current_driver.cylindrical_measured()
    cylindrical_rho = current_driver.rho_measured()

    assert FieldVector(*set_target).is_equal(FieldVector(rho=cylindrical_rho,
                                                         phi=spherical_phi,
                                                         z=cartesian_z))
    assert np.allclose(cylindrical, [cylindrical_rho,
                                     spherical_phi,
                                     cartesian_z])


def get_ramp_down_order(messages: List[str]) -> List[str]:
    order = []

    for msg in messages:
        if "CONF:FIELD:TARG" not in msg:
            continue

        g = re.search(r"\[(.*).*\] Writing: CONF:FIELD:TARG", msg)
        if g is None:
            raise RuntimeError(
                f"No match found in {msg!r} when getting ramp down order")
        name = g.groups()[0]
        order.append(name)

    return order


def test_ramp_down_first(current_driver, caplog):
    """
    To prevent quenching of the magnets, we need the driver to always
    be within the field limits. Part of the strategy of making sure
    that this is the case includes always performing ramp downs
    first. For example, if magnets x and z need to be ramped up, while
    magnet y needs to be ramped down, ramp down magnet y first and
    then ramp up x and z.  This is tested here.
    """
    names = ["x", "y", "z"]
    set_point = np.array([0.3, 0.4, 0.5])
    # Put the magnets in a well defined state
    current_driver.cartesian(set_point)
    # We begin with ramping down x first while ramping up y and z
    delta = np.array([-0.1, 0.1, 0.1])

    log_name = 'qcodes.instrument.base'

    with caplog.at_level(logging.DEBUG, logger=log_name):
        for count, ramp_down_name in enumerate(names):
            # The second iteration will ramp down y while ramping up x and z.
            set_point += np.roll(delta, count)
            # Check if y is adjusted first.
            # We will perform the same test with z in the third iteration

            # clear the message stream
            caplog.clear()

            # make a ramp
            current_driver.cartesian(set_point)
            # get the logging outputs from the instruments.
            messages = [record.message for record in caplog.records]
            # get the order in which the ramps down occur
            order = get_ramp_down_order(messages)
            # the first one should be the one for which delta < 0
            assert order[0][0] == names[count]


def test_field_limit_exception(current_driver):
    """
    Test that an exception is raised if we intentionally set the field
    beyond the limits. Together with the no_test_ramp_down_first test
    this should prevent us from ever exceeding set point limits.  In
    this test we generate a regular grid in three-D and assert that
    the driver can be set to a set point if any of of the requirements
    given by field_limit is satisfied. An error *must* be raised if
    none of the safety limits are satisfied.
    """
    x = np.linspace(-3, 3, 11)
    y = np.copy(x)
    z = np.copy(x)
    set_points = zip(*(i.flatten() for i in np.meshgrid(x, y, z)))

    for set_point in set_points:
        should_not_raise = any([is_safe(*set_point)
                                for is_safe in field_limit])

        if should_not_raise:
            current_driver.cartesian(set_point)
        else:
            with pytest.raises(Exception) as excinfo:
                current_driver.cartesian(set_point)

            assert "field would exceed limit" in excinfo.value.args[0]
            vals_and_setpoints = zip(current_driver.cartesian(), set_point)
            belief = not(all([val == sp for val, sp in vals_and_setpoints]))
            assert belief


def test_cylindrical_poles(current_driver):
    """
    Test that the phi coordinate is remembered even if the resulting
    vector is equivalent to the null vector
    """
    rho, phi, z = 0.4, 30.0, 0.5
    # This is equivalent to the null vector
    current_driver.cylindrical((0.0, phi, 0.0))
    current_driver.rho(rho)
    current_driver.z(z)

    # after setting the rho and z values we should have the vector as
    # originally intended.
    rho_m, phi_m, z_m = current_driver.cylindrical()

    assert np.allclose([rho_m, phi_m, z_m], [rho, phi, z])


def test_spherical_poles(current_driver):
    """
    Test that the theta and phi coordinates are remembered even if the
    resulting vector is equivalent to the null vector
    """

    field, theta, phi = 0.5, 30.0, 50.0
    # If field=0, then this is equivalent to the null vector
    current_driver.spherical((0.0, theta, phi))
    # x, y, z = (0, 0, 0)
    current_driver.field(field)

    # after setting the field we should have the vector as
    # originally intended.
    field_m, theta_m, phi_m = current_driver.spherical()
    assert np.allclose([field_m, theta_m, phi_m], [field, theta, phi])


def test_ramp_rate_exception(current_driver):
    """
    Test that an exception is raised if we try to set the ramp rate
    to a higher value than is allowed
    """
    max_ramp_rate = AMI430_VISA._DEFAULT_CURRENT_RAMP_LIMIT
    target_ramp_rate = max_ramp_rate + 0.01
    ix = current_driver._instrument_x

    with pytest.raises(Exception) as excinfo:
        ix.ramp_rate(target_ramp_rate)

        errmsg = f"must be between 0 and {max_ramp_rate} inclusive"

        assert errmsg in excinfo.value.args[0]


def test_reducing_field_ramp_limit_reduces_a_higher_ramp_rate(ami430):
    """
    When reducing field_ramp_limit, the actual ramp_rate should also be
    reduced if the new field_ramp_limit is lower than the actual ramp_rate
    now.
    """
    factor = 0.8

    # The following fact is expected for the test
    assert ami430.ramp_rate() <= ami430.field_ramp_limit()

    # Set ramp_rate_limit to value that is smaller than the ramp_rate of now
    new_field_ramp_limit = ami430.ramp_rate() * factor
    ami430.field_ramp_limit(new_field_ramp_limit)

    # Assert that the ramp_rate changed to fit within the new field_ramp_limit
    assert ami430.ramp_rate() <= ami430.field_ramp_limit()

    # Well, actually, the new ramp_rate is equal to the new field_ramp_limit
    assert ami430.ramp_rate() == ami430.field_ramp_limit()


def test_reducing_current_ramp_limit_reduces_a_higher_ramp_rate(ami430):
    """
    When reducing current_ramp_limit, the actual ramp_rate should also be
    reduced if the new current_ramp_limit is lower than the actual ramp_rate
    now (with respect to field/current conversion).
    """
    factor = 0.8

    # The following fact is expected for the test
    assert ami430.ramp_rate() \
        <= ami430.current_ramp_limit() * ami430.coil_constant()

    # Set ramp_rate_limit to value that is smaller than the ramp_rate of now
    new_current_ramp_limit = ami430.ramp_rate() \
        * factor / ami430.coil_constant()
    ami430.current_ramp_limit(new_current_ramp_limit)

    # Assert that the ramp_rate changed to fit within the new field_ramp_limit
    assert ami430.ramp_rate() <= ami430.field_ramp_limit()
    assert ami430.ramp_rate() \
        <= ami430.current_ramp_limit() * ami430.coil_constant()

    # Well, actually, the new ramp_rate is equal to the new field_ramp_limit
    assert ami430.ramp_rate() == ami430.field_ramp_limit()


def test_reducing_field_ramp_limit_keeps_a_lower_ramp_rate_as_is(ami430):
    """
    When reducing field_ramp_limit, the actual ramp_rate should remain
    if the new field_ramp_limit is higher than the actual ramp_rate now.
    """
    factor = 1.2

    # The following fact is expected for the test
    assert ami430.ramp_rate() <= ami430.field_ramp_limit()

    old_ramp_rate = ami430.ramp_rate()

    # Set ramp_rate_limit to value that is larger than the ramp_rate of now
    new_field_ramp_limit = ami430.ramp_rate() * factor
    ami430.field_ramp_limit(new_field_ramp_limit)

    # Assert that the ramp_rate remained within the new field_ramp_limit
    assert ami430.ramp_rate() <= ami430.field_ramp_limit()

    # Assert that ramp_rate hasn't actually changed
    assert ami430.ramp_rate() == old_ramp_rate


def test_reducing_current_ramp_limit_keeps_a_lower_ramp_rate_as_is(ami430):
    """
    When reducing current_ramp_limit, the actual ramp_rate should remain
    if the new current_ramp_limit is higher than the actual ramp_rate now
    (with respect to field/current conversion).
    """
    factor = 1.2

    # The following fact is expected for the test
    assert ami430.ramp_rate() \
        <= ami430.current_ramp_limit() * ami430.coil_constant()

    old_ramp_rate = ami430.ramp_rate()

    # Set ramp_rate_limit to value that is larger than the ramp_rate of now
    new_current_ramp_limit = ami430.ramp_rate() \
        * factor / ami430.coil_constant()
    ami430.current_ramp_limit(new_current_ramp_limit)

    # Assert that the ramp_rate remained within the new field_ramp_limit
    assert ami430.ramp_rate() <= ami430.field_ramp_limit()
    assert ami430.ramp_rate() \
        <= ami430.current_ramp_limit() * ami430.coil_constant()

    # Assert that ramp_rate hasn't actually changed
    assert ami430.ramp_rate() == old_ramp_rate


def test_blocking_ramp_parameter(current_driver, caplog):

    assert current_driver.block_during_ramp() is True

    log_name = 'qcodes.instrument.base'

    with caplog.at_level(logging.DEBUG, logger=log_name):
        current_driver.cartesian((0, 0, 0))
        caplog.clear()
        current_driver.cartesian((0, 0, 1))

        messages = [record.message for record in caplog.records]
        assert messages[-1] == '[z(AMI430_VISA)] Finished blocking ramp'
        assert messages[-6] == \
            '[z(AMI430_VISA)] Starting blocking ramp of z to 1.0'

        caplog.clear()
        current_driver.block_during_ramp(False)
        current_driver.cartesian((0, 0, 0))
        messages = [record.message for record in caplog.records]

        assert len([mssg for mssg in messages if 'blocking' in mssg]) == 0


def test_current_and_field_params_interlink_at_init(ami430):
    """
    Test that the values of the ``coil_constant``-dependent parameters
    are correctly proportional to each other at the initialization of the
    instrument driver.
    """
    coil_constant = ami430.coil_constant()
    current_ramp_limit = ami430.current_ramp_limit()
    field_ramp_limit = ami430.field_ramp_limit()
    current_limit = ami430.current_limit()
    field_limit = ami430.field_limit()

    np.testing.assert_almost_equal(
        field_ramp_limit, current_ramp_limit*coil_constant)

    np.testing.assert_almost_equal(
        field_limit, current_limit*coil_constant)


def test_current_and_field_params_interlink__change_current_ramp_limit(
        ami430, factor=0.9):
    """
    Test that after changing ``current_ramp_limit``, the values of the
    ``field_*`` parameters change proportionally, ``coil__constant`` remains
    the same. At the end just ensure that the values of the
    ``coil_constant``-dependent parameters are correctly proportional to each
    other.
    """
    coil_constant_old = ami430.coil_constant()
    current_ramp_limit_old = ami430.current_ramp_limit()
    field_ramp_limit_old = ami430.field_ramp_limit()
    current_limit_old = ami430.current_limit()
    field_limit_old = ami430.field_limit()

    current_ramp_limit_new = current_ramp_limit_old * factor

    ami430.current_ramp_limit(current_ramp_limit_new)

    field_ramp_limit_new_expected = field_ramp_limit_old * factor

    current_ramp_limit = ami430.current_ramp_limit()
    field_ramp_limit = ami430.field_ramp_limit()
    coil_constant = ami430.coil_constant()
    current_limit = ami430.current_limit()
    field_limit = ami430.field_limit()

    # The following parameters are expected to change
    np.testing.assert_almost_equal(
        current_ramp_limit, current_ramp_limit_new)
    np.testing.assert_almost_equal(
        field_ramp_limit, field_ramp_limit_new_expected)

    # The following parameters are not expected to change
    np.testing.assert_almost_equal(
        coil_constant, coil_constant_old)
    np.testing.assert_almost_equal(
        current_limit, current_limit_old)
    np.testing.assert_almost_equal(
        field_limit, field_limit_old)

    # Proportions are expected to hold between field and current parameters
    np.testing.assert_almost_equal(
        field_ramp_limit, current_ramp_limit*coil_constant)
    np.testing.assert_almost_equal(
        field_limit, current_limit*coil_constant)


def test_current_and_field_params_interlink__change_field_ramp_limit(
        ami430, factor=0.9):
    """
    Test that after changing ``field_ramp_limit``, the values of the
    ``current_*`` parameters change proportionally, ``coil__constant`` remains
    the same. At the end just ensure that the values of the
    ``coil_constant``-dependent parameters are correctly proportional to each
    other.
    """
    coil_constant_old = ami430.coil_constant()
    current_ramp_limit_old = ami430.current_ramp_limit()
    field_ramp_limit_old = ami430.field_ramp_limit()
    current_limit_old = ami430.current_limit()
    field_limit_old = ami430.field_limit()

    field_ramp_limit_new = field_ramp_limit_old * factor

    ami430.field_ramp_limit(field_ramp_limit_new)

    current_ramp_limit_new_expected = current_ramp_limit_old * factor

    current_ramp_limit = ami430.current_ramp_limit()
    field_ramp_limit = ami430.field_ramp_limit()
    coil_constant = ami430.coil_constant()
    current_limit = ami430.current_limit()
    field_limit = ami430.field_limit()

    # The following parameters are expected to change
    np.testing.assert_almost_equal(
        field_ramp_limit, field_ramp_limit_new)
    np.testing.assert_almost_equal(
        current_ramp_limit, current_ramp_limit_new_expected)

    # The following parameters are not expected to change
    np.testing.assert_almost_equal(
        coil_constant, coil_constant_old)
    np.testing.assert_almost_equal(
        current_limit, current_limit_old)
    np.testing.assert_almost_equal(
        field_limit, field_limit_old)

    # Proportions are expected to hold between field and current parameters
    np.testing.assert_almost_equal(
        field_ramp_limit, current_ramp_limit*coil_constant)
    np.testing.assert_almost_equal(
        field_limit, current_limit*coil_constant)


def test_current_and_field_params_interlink__change_coil_constant(
        ami430, factor=3):
    """
    Test that after changing ``change_coil_constant``, the values of the
    ``current_*`` parameters remain the same while the values of the
    ``field_*`` parameters change proportionally. At the end just ensure that
    the values of the ``coil_constant``-dependent parameters are correctly
    proportional to each other.
    """
    coil_constant_old = ami430.coil_constant()
    current_ramp_limit_old = ami430.current_ramp_limit()
    field_ramp_limit_old = ami430.field_ramp_limit()
    current_limit_old = ami430.current_limit()
    field_limit_old = ami430.field_limit()

    coil_constant_new = coil_constant_old * factor

    ami430.coil_constant(coil_constant_new)

    current_ramp_limit_new_expected = current_ramp_limit_old
    current_limit_new_expected = current_limit_old
    field_ramp_limit_new_expected = field_ramp_limit_old * factor
    field_limit_new_expected = field_limit_old * factor

    current_ramp_limit = ami430.current_ramp_limit()
    field_ramp_limit = ami430.field_ramp_limit()
    coil_constant = ami430.coil_constant()
    current_limit = ami430.current_limit()
    field_limit = ami430.field_limit()

    # The following parameters are expected to change
    np.testing.assert_almost_equal(
        coil_constant, coil_constant_new)
    np.testing.assert_almost_equal(
        current_ramp_limit, current_ramp_limit_new_expected)
    np.testing.assert_almost_equal(
        field_ramp_limit, field_ramp_limit_new_expected)
    np.testing.assert_almost_equal(
        current_limit, current_limit_new_expected)
    np.testing.assert_almost_equal(
        field_limit, field_limit_new_expected)

    # Proportions are expected to hold between field and current parameters
    np.testing.assert_almost_equal(
        field_ramp_limit, current_ramp_limit*coil_constant)
    np.testing.assert_almost_equal(
        field_limit, current_limit*coil_constant)


def test_current_and_field_params_interlink__permutations_of_tests(ami430):
    """
    As per one of the user's request, the
    test_current_and_field_params_interlink__* tests are executed here with
    arbitrary 'factor's and with all permutations. This test ensures the
    robustness of the driver even more.

    Note that the 'factor's are randomized "manually" because of the
    possibility to hit the limits of the parameters.
    """
    with warnings.catch_warnings():
        # this is to avoid AMI430Warning about "maximum ramp rate", which
        # may show up but is not relevant to this test
        warnings.simplefilter('ignore', category=AMI430Warning)

        test_current_and_field_params_interlink_at_init(ami430)

        test_current_and_field_params_interlink__change_coil_constant(
            ami430, factor=1.2)
        test_current_and_field_params_interlink__change_field_ramp_limit(
            ami430, factor=1.0023)
        test_current_and_field_params_interlink__change_current_ramp_limit(
            ami430, factor=0.98)

        test_current_and_field_params_interlink__change_coil_constant(
            ami430, factor=1.53)
        test_current_and_field_params_interlink__change_current_ramp_limit(
            ami430, factor=2.0)
        test_current_and_field_params_interlink__change_field_ramp_limit(
            ami430, factor=0.633)

        test_current_and_field_params_interlink__change_field_ramp_limit(
            ami430, factor=1.753)
        test_current_and_field_params_interlink__change_coil_constant(
            ami430, factor=0.876)
        test_current_and_field_params_interlink__change_current_ramp_limit(
            ami430, factor=4.6)

        test_current_and_field_params_interlink__change_coil_constant(
            ami430, factor=1.87)
        test_current_and_field_params_interlink__change_current_ramp_limit(
            ami430, factor=2.11)
        test_current_and_field_params_interlink__change_field_ramp_limit(
            ami430, factor=1.0020)

        test_current_and_field_params_interlink__change_field_ramp_limit(
            ami430, factor=0.42)
        test_current_and_field_params_interlink__change_current_ramp_limit(
            ami430, factor=3.1415)
        test_current_and_field_params_interlink__change_coil_constant(
            ami430, factor=1.544)

        test_current_and_field_params_interlink__change_coil_constant(
            ami430, factor=0.12)
        test_current_and_field_params_interlink__change_field_ramp_limit(
            ami430, factor=0.4422)
        test_current_and_field_params_interlink__change_current_ramp_limit(
            ami430, factor=0.00111)


def _parametrization_kwargs():
    kwargs = {'argvalues': [], 'ids': []}

    for type_constructor, type_name in zip(
        ((int, float)
         + numpy_concrete_ints
         + numpy_non_concrete_ints_instantiable
         + numpy_concrete_floats
         + numpy_non_concrete_floats_instantiable),
        (['int', 'float']
         + [str(t) for t in numpy_concrete_ints]
         + [str(t) for t in numpy_non_concrete_ints_instantiable]
         + [str(t) for t in numpy_concrete_floats]
         + [str(t) for t in numpy_non_concrete_floats_instantiable])
    ):
        kwargs['argvalues'].append(type_constructor(2.2))
        kwargs['ids'].append(type_name)

    return kwargs


@pytest.mark.parametrize('field_limit', **_parametrization_kwargs())
def test_numeric_field_limit(magnet_axes_instances, field_limit, request):
    mag_x, mag_y, mag_z = magnet_axes_instances
    ami = AMI430_3D("AMI430-3D", mag_x, mag_y, mag_z, field_limit)
    request.addfinalizer(ami.close)

    assert isinstance(ami._field_limit, float)

    target_within_limit = (field_limit * 0.95, 0, 0)
    ami.cartesian(target_within_limit)

    target_outside_limit = (field_limit * 1.05, 0, 0)
    with pytest.raises(ValueError,
                       match='_set_fields aborted; field would exceed limit'):
        ami.cartesian(target_outside_limit)


def test_ramp_rate_units_and_field_units_at_init(ami430):
    """
    Test values of ramp_rate_units and field_units parameters at init,
    and the units of other parameters which depend on the
    values of ramp_rate_units and field_units parameters.
    """
    initial_ramp_rate_units = ami430.ramp_rate_units()
    initial_field_units = ami430.field_units()

    assert initial_ramp_rate_units == 'seconds'
    assert initial_field_units == 'tesla'

    assert ami430.coil_constant.unit == "T/A"
    assert ami430.field_limit.unit == "T"
    assert ami430.field.unit == "T"
    assert ami430.setpoint.unit == "T"
    assert ami430.ramp_rate.unit == "T/s"
    assert ami430.current_ramp_limit.unit == "A/s"
    assert ami430.field_ramp_limit.unit == "T/s"


@pytest.mark.parametrize(('new_value', 'unit_string', 'scale'),
                         (('seconds', 's', 1), ('minutes', 'min', 1/60)),
                         ids=('seconds', 'minutes'))
def test_change_ramp_rate_units_parameter(ami430, new_value, unit_string,
                                          scale):
    """
    Test that changing value of ramp_rate_units parameter is reflected in
    settings of other magnet parameters.
    """
    coil_constant_unit = ami430.coil_constant.unit
    field_limit_unit = ami430.field_limit.unit
    field_unit = ami430.field.unit
    setpoint_unit = ami430.setpoint.unit
    coil_constant_timestamp = ami430.coil_constant.get_latest.get_timestamp()
    # this prevents possible flakiness of the timestamp comparison
    # later in the test that may originate from the not-enough resolution
    # of the time function used in `Parameter` and `GetLatest` classes
    time.sleep(2 * _time_resolution)

    ami430.ramp_rate_units(new_value)

    ramp_rate_units__actual = ami430.ramp_rate_units()
    assert ramp_rate_units__actual == new_value

    assert ami430.coil_constant.unit == coil_constant_unit
    assert ami430.field_limit.unit == field_limit_unit
    assert ami430.field.unit == field_unit
    assert ami430.setpoint.unit == setpoint_unit

    assert ami430.ramp_rate.unit.endswith("/" + unit_string)
    assert ami430.current_ramp_limit.unit.endswith("/" + unit_string)
    assert ami430.field_ramp_limit.unit.endswith("/" + unit_string)

    assert ami430.current_ramp_limit.scale == scale

    # Assert `coil_constant` value has been updated
    assert ami430.coil_constant.get_latest.get_timestamp() \
           > coil_constant_timestamp


@pytest.mark.parametrize(('new_value', 'unit_string'),
                         (('tesla', 'T'), ('kilogauss', 'kG')),
                         ids=('tesla', 'kilogauss'))
def test_change_field_units_parameter(ami430, new_value, unit_string):
    """
    Test that changing value of field_units parameter is reflected in
    settings of other magnet parameters.
    """
    current_ramp_limit_unit = ami430.current_ramp_limit.unit
    current_ramp_limit_scale = ami430.current_ramp_limit.scale
    coil_constant_timestamp = ami430.coil_constant.get_latest.get_timestamp()
    # this prevents possible flakiness of the timestamp comparison
    # later in the test that may originate from the not-enough resolution
    # of the time function used in `Parameter` and `GetLatest` classes
    time.sleep(2 * _time_resolution)

    ami430.field_units(new_value)

    field_units__actual = ami430.field_units()
    assert field_units__actual == new_value

    assert ami430.current_ramp_limit.unit == current_ramp_limit_unit
    assert ami430.current_ramp_limit.scale == current_ramp_limit_scale

    assert ami430.field_limit.unit == unit_string
    assert ami430.field.unit == unit_string
    assert ami430.setpoint.unit == unit_string

    assert ami430.coil_constant.unit.startswith(unit_string + "/")
    assert ami430.ramp_rate.unit.startswith(unit_string + "/")
    assert ami430.field_ramp_limit.unit.startswith(unit_string + "/")

    # Assert `coil_constant` value has been updated
    assert ami430.coil_constant.get_latest.get_timestamp() \
           > coil_constant_timestamp



from copy import deepcopy
from typing import Any, List, Optional, Sequence, Union

from typing_extensions import TypedDict


class ParamSpecBaseDict(TypedDict):
    name: str
    paramtype: str
    label: Optional[str]
    unit: Optional[str]


class ParamSpecDict(ParamSpecBaseDict):
    inferred_from: List[str]
    depends_on: List[str]


class ParamSpecBase:

    allowed_types = ['array', 'numeric', 'text', 'complex']

    def __init__(self,
                 name: str,
                 paramtype: str,
                 label: Optional[str] = None,
                 unit: Optional[str] = None):
        """
        Args:
            name: name of the parameter
            paramtype: type of the parameter, i.e. the SQL storage class
            label: label of the parameter
            unit: The unit of the parameter
        """

        if not isinstance(paramtype, str):
            raise ValueError('Paramtype must be a string.')
        if paramtype.lower() not in self.allowed_types:
            raise ValueError("Illegal paramtype. Must be on of "
                             f"{self.allowed_types}")
        if not name.isidentifier():
            raise ValueError(f'Invalid name: {name}. Only valid python '
                             'identifier names are allowed (no spaces or '
                             'punctuation marks, no prepended '
                             'numbers, etc.)')

        self.name = name
        self.type = paramtype.lower()
        self.label = label or ''
        self.unit = unit or ''

        self._hash: int = self._compute_hash()

    def _compute_hash(self) -> int:
        """
        This method should only be called by __init__
        """
        attrs = ['name', 'type', 'label', 'unit']
        # First, get the hash of the tuple with all the relevant attributes
        all_attr_tuple_hash = hash(
            tuple(getattr(self, attr) for attr in attrs)
        )
        hash_value = all_attr_tuple_hash

        # Then, XOR it with the individual hashes of all relevant attributes
        for attr in attrs:
            hash_value = hash_value ^ hash(getattr(self, attr))

        return hash_value

    def sql_repr(self) -> str:
        return f"{self.name} {self.type}"

    def __repr__(self) -> str:
        return (f"ParamSpecBase('{self.name}', '{self.type}', '{self.label}', "
                f"'{self.unit}')")

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ParamSpecBase):
            return False
        attrs = ['name', 'type', 'label', 'unit']
        for attr in attrs:
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True

    def __hash__(self) -> int:
        """
        Allow ParamSpecBases in data structures that use hashing (e.g. sets)
        """
        return self._hash

    def _to_dict(self) -> ParamSpecBaseDict:
        """
        Write the ParamSpec as a dictionary
        """
        output = ParamSpecBaseDict(name=self.name,
                                   paramtype=self.type,
                                   label=self.label,
                                   unit=self.unit)
        return output

    @classmethod
    def _from_dict(cls, ser: ParamSpecBaseDict) -> 'ParamSpecBase':
        """
        Create a ParamSpec instance of the current version
        from a dictionary representation of ParamSpec of some version

        The version changes must be implemented as a series of transformations
        of the representation dict.
        """

        return ParamSpecBase(name=ser['name'],
                             paramtype=ser['paramtype'],
                             label=ser['label'],
                             unit=ser['unit'])


class ParamSpec(ParamSpecBase):

    def __init__(self, name: str,
                 paramtype: str,
                 label: Optional[str] = None,
                 unit: Optional[str] = None,
                 inferred_from: Optional[Sequence[Union['ParamSpec', str]]] = None,
                 depends_on: Optional[Sequence[Union['ParamSpec', str]]] = None,
                 **metadata: Any) -> None:
        """
        Args:
            name: name of the parameter
            paramtype: type of the parameter, i.e. the SQL storage class
            label: label of the parameter
            inferred_from: the parameters that this parameter is inferred from
            depends_on: the parameters that this parameter depends on
        """

        super().__init__(name, paramtype, label, unit)

        self._inferred_from: List[str] = []
        self._depends_on: List[str] = []

        inferred_from = [] if inferred_from is None else inferred_from
        depends_on = [] if depends_on is None else depends_on

        if isinstance(inferred_from, str):
            raise ValueError(f"ParamSpec {self.name} got "
                             f"string {inferred_from} as inferred_from. "
                             f"It needs a "
                             f"Sequence of ParamSpecs or strings")
        self._inferred_from.extend(
            p.name if isinstance(p, ParamSpec) else p
            for p in inferred_from)

        if isinstance(depends_on, str):
            raise ValueError(f"ParamSpec {self.name} got "
                             f"string {depends_on} as depends_on. It needs a "
                             f"Sequence of ParamSpecs or strings")
        self._depends_on.extend(
            p.name if isinstance(p, ParamSpec) else p
            for p in depends_on)

        if metadata:
            self.metadata = metadata

    @property
    def inferred_from_(self) -> List[str]:
        return deepcopy(self._inferred_from)

    @property
    def depends_on_(self) -> List[str]:
        return deepcopy(self._depends_on)

    @property
    def inferred_from(self) -> str:
        return ', '.join(self._inferred_from)

    @property
    def depends_on(self) -> str:
        return ', '.join(self._depends_on)

    def copy(self) -> 'ParamSpec':
        """
        Make a copy of self
        """
        return ParamSpec(self.name, self.type, self.label, self.unit,
                         deepcopy(self._inferred_from),
                         deepcopy(self._depends_on))

    def __repr__(self) -> str:
        return (f"ParamSpec('{self.name}', '{self.type}', '{self.label}', "
                f"'{self.unit}', inferred_from={self._inferred_from}, "
                f"depends_on={self._depends_on})")

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ParamSpec):
            return False
        string_attrs = ['name', 'type', 'label', 'unit']
        list_attrs = ['_inferred_from', '_depends_on']
        for string_attr in string_attrs:
            if getattr(self, string_attr) != getattr(other, string_attr):
                return False
        for list_attr in list_attrs:
            ours = getattr(self, list_attr)
            theirs = getattr(other, list_attr)
            if ours != theirs:
                return False
        return True

    def __hash__(self) -> int:
        """Allow ParamSpecs in data structures that use hashing (i.e. sets)"""
        attrs_with_strings = ['name', 'type', 'label', 'unit']
        attrs_with_lists = ['_inferred_from', '_depends_on']

        # First, get the hash of the tuple with all the relevant attributes
        all_attr_tuple_hash = hash(
            tuple(getattr(self, attr) for attr in attrs_with_strings)
            + tuple(tuple(getattr(self, attr)) for attr in attrs_with_lists)
        )
        hash_value = all_attr_tuple_hash

        # Then, XOR it with the individual hashes of all relevant attributes
        for attr in attrs_with_strings:
            hash_value = hash_value ^ hash(getattr(self, attr))
        for attr in attrs_with_lists:
            hash_value = hash_value ^ hash(tuple(getattr(self, attr)))

        return hash_value

    def _to_dict(self) -> ParamSpecDict:
        """
        Write the ParamSpec as a dictionary
        """
        basedict = super()._to_dict()
        output = ParamSpecDict(name=basedict['name'],
                               paramtype=basedict['paramtype'],
                               label=basedict['label'],
                               unit=basedict['unit'],
                               inferred_from=self._inferred_from,
                               depends_on=self._depends_on)
        return output

    def base_version(self) -> ParamSpecBase:
        """
        Return a ParamSpecBase object with the same name, paramtype, label
        and unit as this ParamSpec
        """
        return ParamSpecBase(name=self.name,
                             paramtype=self.type,
                             label=self.label,
                             unit=self.unit)

    @classmethod
    def _from_dict(  # type: ignore[override]
            cls,
            ser: ParamSpecDict
    ) -> 'ParamSpec':
        """
        Create a ParamSpec instance of the current version
        from a dictionary representation of ParamSpec of some version

        The version changes must be implemented as a series of transformations
        of the representation dict.
        """

        return ParamSpec(name=ser['name'],
                         paramtype=ser['paramtype'],
                         label=ser['label'],
                         unit=ser['unit'],
                         inferred_from=ser['inferred_from'],
                         depends_on=ser['depends_on'])



# pylint: disable=redefined-outer-name
import pytest
import logging

from qcodes.instrument_drivers.Keysight.keysight_34980a import Keysight34980A
import qcodes.instrument.sims as sims

VISALIB = sims.__file__.replace('__init__.py', 'keysight_34980A.yaml@sim')


@pytest.fixture(scope="module")
def switch_driver():
    inst = Keysight34980A('keysight_34980A_sim',
                          address='GPIB::1::INSTR',
                          visalib=VISALIB)

    try:
        yield inst
    finally:
        inst.close()


def test_safety_interlock_during_init(switch_driver, caplog):
    """
    to check if a warning would show when initialize the instrument with a
    module in safety interlock state. This test has to be placed first if
    the scope is set to be "module".
    """
    msg = [
        x.message for x in caplog.get_records('setup')
        if x.levelno == logging.WARNING
    ]
    assert "safety interlock" in msg[0]


def test_get_idn(switch_driver):
    """
    to check if the instrument attributes are set correctly after getting
    the IDN
    """
    assert switch_driver.IDN() == {
        "vendor": "Keysight",
        "model": "34980A",
        "serial": "1000",
        "firmware": "0.1"
    }


def test_scan_slots(switch_driver):
    """
    to check if the submodule attributes are set correctly after scanning
    every slot
    """
    assert len(switch_driver.system_slots_info) == 2

    assert switch_driver.system_slots_info[1] == {
        "vendor": "Agilent Technologies",
        "model": "34934A-8x64",
        "serial": "AB10000000",
        "firmware": "1.00"
    }

    assert switch_driver.system_slots_info[3] == {
        "vendor": "Agilent Technologies",
        "model": "34934A-4x32",
        "serial": "AB10000001",
        "firmware": "1.00"
    }


def test_safety_interlock(switch_driver, caplog):
    """
    to check if a warning would show when talk to a module that is in safety
    interlock state
    """
    switch_driver.module[3].write('*CLS')
    with caplog.at_level(logging.DEBUG):
        assert "safety interlock" in caplog.text



import math
from unittest.mock import MagicMock

from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1517A import \
    B1517A
from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1500_module import \
    parse_module_query_response, format_dcorr_response, _DCORRResponse, \
    fixed_negative_float, get_name_label_unit_of_impedance_model, \
    convert_dummy_val_to_nan, _FMTResponse, \
    convert_dummy_val_to_nan
from qcodes.instrument_drivers.Keysight.keysightb1500.constants import \
    SlotNr, DCORR, IMP


def test_is_enabled():
    mainframe = MagicMock()

    # Use concrete subclass because B1500Module does not assign channels
    smu = B1517A(parent=mainframe, name='B1517A', slot_nr=1)

    mainframe.ask.return_value = 'CN 1,2,4,8'
    assert smu.is_enabled()
    mainframe.ask.assert_called_once_with('*LRN? 0')

    mainframe.reset_mock(return_value=True)
    mainframe.ask.return_value = 'CN 2,4,8'
    assert not smu.is_enabled()
    mainframe.ask.assert_called_once_with('*LRN? 0')

    mainframe.reset_mock(return_value=True)
    mainframe.ask.return_value = 'CN'
    assert not smu.is_enabled()
    mainframe.ask.assert_called_once_with('*LRN? 0')


def test_enable_outputs():
    mainframe = MagicMock()

    slot_nr = 1
    # Use concrete subclass because B1500Module does not assign channels
    smu = B1517A(parent=mainframe, name='B1517A', slot_nr=slot_nr)

    smu.enable_outputs()
    mainframe.write.assert_called_once_with(f'CN {slot_nr}')


def test_disable_outputs():
    mainframe = MagicMock()

    slot_nr = 1
    # Use concrete subclass because B1500Module does not assign channels
    smu = B1517A(parent=mainframe, name='B1517A', slot_nr=slot_nr)

    smu.disable_outputs()
    mainframe.write.assert_called_once_with(f'CL {slot_nr}')


def test_parse_module_query_response():
    response = 'B1517A,0;B1517A,0;B1520A,0;0,0;0,0;0,0;0,0;0,0;0,0;0,0'
    expected = {SlotNr.SLOT01: 'B1517A',
                SlotNr.SLOT02: 'B1517A',
                SlotNr.SLOT03: 'B1520A'}

    actual = parse_module_query_response(response)

    assert actual == expected


def test_format_dcorr_response():
    resp_str1 = format_dcorr_response(
        _DCORRResponse(mode=DCORR.Mode.Cp_G, primary=0.001, secondary=0.34))
    assert resp_str1 == 'Mode: Cp_G, Primary Cp: 0.001 F, Secondary G: 0.34 S'

    resp_str2 = format_dcorr_response(
        _DCORRResponse(mode=DCORR.Mode.Ls_Rs, primary=0.2, secondary=3.0))
    assert resp_str2 == 'Mode: Ls_Rs, Primary Ls: 0.2 H, Secondary Rs: 3.0 Ω'


def test_fixed_negative_float():
    assert fixed_negative_float('-0.-1') == -0.1
    assert fixed_negative_float('-1.-1') == -1.1
    assert fixed_negative_float('0.1') == 0.1
    assert fixed_negative_float('1.0') == 1.0
    assert fixed_negative_float('1') == 1.0
    assert fixed_negative_float('-1') == -1.0


def test_get_name_label_unit_of_impedance_model():
    model = IMP.MeasurementMode.Cp_D
    name, label, unit = get_name_label_unit_of_impedance_model(model)
    assert name == ('parallel_capacitance', 'dissipation_factor')
    assert label == ('Parallel Capacitance', 'Dissipation Factor')
    assert unit == ('F', '')

    model = IMP.MeasurementMode.Y_THETA_DEG
    name, label, unit = get_name_label_unit_of_impedance_model(model)
    assert name == ('admittance', 'phase')
    assert label == ('Admittance', 'Phase')
    assert unit == ('S', 'degree')


def test_convert_dummy_val_to_nan():
    status = ['C', 'V', 'N', 'V', 'N', 'N']
    value = [0, 199.999e99, 1, 199.999e99, 2, 3]
    channel = [1, 1, 1, 1, 1, 1]
    param_type = ['V', 'V', 'V', 'V', 'V', 'V']
    param = _FMTResponse(value, status, channel, param_type)
    convert_dummy_val_to_nan(param)
    assert math.isnan(param.value[1])
    assert math.isnan(param.value[3])



"""
Test suite for instument.base.*
"""
import contextlib
import io
import re
import weakref

import pytest

from qcodes.instrument.base import Instrument, InstrumentBase, find_or_create_instrument
from qcodes.instrument.function import Function
from qcodes.instrument.parameter import Parameter

from .instrument_mocks import DummyInstrument, MockMetaParabola, MockParabola


@pytest.fixture(name='testdummy', scope='function')
def _dummy_dac():
    instrument = DummyInstrument(
        name='testdummy', gates=['dac1', 'dac2', 'dac3'])
    try:
        yield instrument
    finally:
        instrument.close()


@pytest.fixture(name='parabola', scope='function')
def _dummy_parabola():
    instrument = MockParabola("parabola")
    try:
        yield instrument
    finally:
        instrument.close()


@pytest.fixture(name='close_before_and_after', scope='function')
def _close_before_and_after():
    Instrument.close_all()
    try:
        yield
    finally:
        Instrument.close_all()


def test_validate_function(testdummy):
    testdummy.validate_status()  # test the instrument has valid values

    testdummy.dac1.cache._value = 1000  # overrule the validator
    testdummy.dac1.cache._raw_value = 1000  # overrule the validator
    with pytest.raises(Exception):
        testdummy.validate_status()


def test_check_instances(testdummy):
    with pytest.raises(KeyError, match='Another instrument has the name: testdummy'):
        DummyInstrument(name='testdummy', gates=['dac1', 'dac2', 'dac3'])

    assert Instrument.instances() == []
    assert DummyInstrument.instances() == [testdummy]
    assert testdummy.instances() == [testdummy]


def test_attr_access(testdummy):

    # test the instrument works
    testdummy.dac1.set(10)
    val = testdummy.dac1.get()
    assert val == 10

    # close the instrument
    testdummy.close()

    # make sure the name property still exists
    assert hasattr(testdummy, 'name')
    assert testdummy.name == 'testdummy'

    # make sure we can still print the instrument
    assert 'testdummy' in testdummy.__repr__()
    assert 'testdummy' in str(testdummy)

    # make sure the gate is removed
    assert not hasattr(testdummy, 'dac1')


def test_get_idn(testdummy):
    idn = dict(zip(('vendor', 'model', 'serial', 'firmware'),
                   [None, testdummy.name, None, None]))
    assert testdummy.get_idn() == idn


def test_repr(testdummy):
    assert repr(testdummy) == '<DummyInstrument: testdummy>'


def test_add_remove_f_p(testdummy):
    with pytest.raises(KeyError, match="Duplicate parameter name dac1"):
        testdummy.add_parameter("dac1", get_cmd="foo")

    testdummy.add_function('function', call_cmd='foo')

    with pytest.raises(KeyError, match='Duplicate function name function'):
        testdummy.add_function('function', call_cmd='foo')

    testdummy.add_function('dac1', call_cmd='foo')

    # test custom __get_attr__ for functions
    fcn = testdummy['function']
    assert isinstance(fcn, Function)
    # by design, one gets the parameter if a function exists
    # and has same name
    dac1 = testdummy['dac1']
    assert isinstance(dac1, Parameter)


def test_instances(testdummy, parabola):
    instruments = [testdummy, parabola]
    for instrument in instruments:
        for other_instrument in instruments:
            instances = instrument.instances()
            # check that each instrument is in only its own
            if other_instrument is instrument:
                assert instrument in instances
            else:
                assert other_instrument not in instances

            # check that we can find each instrument from any other
            assert instrument is other_instrument.find_instrument(instrument.name)

        # check that we can find this instrument from the base class
        assert instrument is Instrument.find_instrument(instrument.name)


def test_is_valid(testdummy):
    assert Instrument.is_valid(testdummy)
    testdummy.close()
    assert not Instrument.is_valid(testdummy)


def test_snapshot_value(testdummy):
    testdummy.add_parameter('has_snapshot_value',
                            parameter_class=Parameter,
                            initial_value=42,
                            snapshot_value=True,
                            get_cmd=None, set_cmd=None)
    testdummy.add_parameter('no_snapshot_value',
                            parameter_class=Parameter,
                            initial_value=42,
                            snapshot_value=False,
                            get_cmd=None, set_cmd=None)

    snapshot = testdummy.snapshot()

    assert 'name' in snapshot
    assert 'testdummy' in snapshot['name']

    assert 'value' in snapshot['parameters']['has_snapshot_value']
    assert 42 == snapshot['parameters']['has_snapshot_value']['value']
    assert 'value' not in snapshot['parameters']['no_snapshot_value']


def test_meta_instrument(parabola):
    mock_instrument = MockMetaParabola("mock_parabola", parabola)

    # Check that the mock instrument can return values
    assert mock_instrument.parabola() == parabola.parabola()
    mock_instrument.x(1)
    mock_instrument.y(2)
    assert mock_instrument.parabola() == parabola.parabola()
    assert mock_instrument.parabola() != 0

    # Add a scaling factor
    mock_instrument.gain(2)
    assert mock_instrument.parabola() == parabola.parabola()*2

    # Check snapshots
    snap = mock_instrument.snapshot(update=True)
    assert "parameters" in snap
    assert "gain" in snap["parameters"]
    assert snap["parameters"]["gain"]["value"] == 2

    # Check printable snapshot
    f = io.StringIO()
    with contextlib.redirect_stdout(f):
        mock_instrument.print_readable_snapshot()
    readable_snap = f.getvalue()

    # Line length satisfied
    assert all(len(line) <= 80 for line in readable_snap.splitlines())
    # Gain is included in output with correct value
    assert re.search(r"gain[ \t]+:[ \t]+2", readable_snap) is not None


def test_find(testdummy):
    """Test finding an existing instrument"""

    instr_2 = find_or_create_instrument(
        DummyInstrument, name='testdummy', gates=['dac1', 'dac2', 'dac3'])

    assert instr_2 is testdummy
    assert instr_2.name == testdummy.name


def test_find_same_name_but_different_class(close_before_and_after, request):
    """Test finding an existing instrument with different class"""
    instr = DummyInstrument(
        name='instr', gates=['dac1', 'dac2', 'dac3'])
    request.addfinalizer(instr.close)

    class GammyInstrument(Instrument):
        some_other_attr = 25

    # Find an instrument with the same name but of different class
    error_msg = ("Instrument instr is <class "
                 "'qcodes.tests.instrument_mocks.DummyInstrument'> but "
                 "<class 'qcodes.tests.test_instrument"
                 ".test_find_same_name_but_different_class.<locals>"
                 ".GammyInstrument'> was requested")

    with pytest.raises(TypeError, match=error_msg):
        _ = find_or_create_instrument(
            GammyInstrument, name='instr', gates=['dac1', 'dac2', 'dac3'])


def test_create(close_before_and_after, request):
    """Test creating an instrument that does not yet exist"""
    instr = find_or_create_instrument(
        DummyInstrument, name='instr', gates=['dac1', 'dac2', 'dac3'])
    request.addfinalizer(instr.close)
    assert 'instr' == instr.name


def test_other_exception(close_before_and_after):
    """Test an unexpected exception occurred during finding instrument"""
    with pytest.raises(TypeError, match="unhashable type: 'dict'"):
        # in order to raise an unexpected exception, and make sure it is
        # passed through the call stack, let's pass an empty dict instead
        # of a string with instrument name
        _ = find_or_create_instrument(DummyInstrument, {})


def test_recreate(close_before_and_after, request):
    """Test the case when instrument needs to be recreated"""
    instr = DummyInstrument(
        name='instr', gates=['dac1', 'dac2', 'dac3'])
    request.addfinalizer(instr.close)
    instr_ref = weakref.ref(instr)

    assert ['instr'] == list(Instrument._all_instruments.keys())

    instr_2 = find_or_create_instrument(
        DummyInstrument, name='instr', gates=['dac1', 'dac2'],
        recreate=True
    )
    request.addfinalizer(instr_2.close)
    instr_2_ref = weakref.ref(instr_2)

    assert ['instr'] == list(Instrument._all_instruments.keys())

    assert instr_2_ref in Instrument._all_instruments.values()
    assert instr_ref not in Instrument._all_instruments.values()


def test_instrument_metadata(request):
    metadatadict = {1: "data", "some": "data"}
    instrument = DummyInstrument(name='testdummy', gates=['dac1', 'dac2', 'dac3'],
                                 metadata=metadatadict)
    request.addfinalizer(instrument.close)
    assert instrument.metadata == metadatadict


def test_instrumentbase_metadata():
    metadatadict = {1: "data", "some": "data"}
    instrument = InstrumentBase('instr', metadata=metadatadict)
    assert instrument.metadata == metadatadict


def test_snapshot_and_meta_attrs():
    """Test snapshot of InstrumentBase contains _meta_attrs attributes"""
    instr = InstrumentBase('instr')

    assert instr.name == 'instr'

    assert hasattr(instr, '_meta_attrs')
    assert instr._meta_attrs == ['name']

    snapshot = instr.snapshot()

    assert 'name' in snapshot
    assert 'instr' == snapshot['name']

    assert '__class__' in snapshot
    assert 'InstrumentBase' in snapshot['__class__']



from .base import Instrument, find_or_create_instrument
from .ip import IPInstrument
from .visa import VisaInstrument
from .channel import InstrumentChannel, ChannelList
from .function import Function
from .parameter import (
    Parameter,
    ArrayParameter,
    MultiParameter,
    ParameterWithSetpoints,
    DelegateParameter,
    ManualParameter,
    ScaledParameter,
    combine,
    CombinedParameter)
from .sweep_values import SweepFixedValues, SweepValues



from enum import IntFlag
from typing import Dict, cast, List, Tuple, Iterable, TextIO, Any, Optional
from itertools import takewhile

from qcodes import VisaInstrument, InstrumentChannel, ChannelList
from qcodes.utils.validators import Enum, Numbers
from qcodes.instrument.group_parameter import GroupParameter, Group


def read_curve_file(curve_file: TextIO) -> Dict[Any, Any]:
    """
    Read a curve file with extension .330
    The file format of this file is shown in test_lakeshore_file_parser.py
    in the test module

    The output is a dictionary with keys: "metadata" and "data".
    The metadata dictionary contains the first n lines of the curve file which
    are in the format "item: value". The data dictionary contains the actual
    curve data.
    """

    def split_data_line(line: str, parser: type = str) -> List[Any]:
        return [parser(i) for i in line.split("  ") if i != ""]

    def strip(strings: Iterable[str]) -> Tuple[str, ...]:
        return tuple(s.strip() for s in strings)

    lines = iter(curve_file.readlines())
    # Meta data lines contain a colon
    metadata_lines = takewhile(lambda s: ":" in s, lines)
    # Data from the file is collected in the following dict
    file_data: Dict[str, Dict[str, Any]] = dict()
    # Capture meta data
    parsed_lines = [strip(line.split(":")) for line in metadata_lines]
    file_data["metadata"] = {key: value for key, value in parsed_lines}
    # After meta data we have a data header
    header_items = strip(split_data_line(next(lines)))
    # After that we have the curve data
    data: List[List[float]] = [
        split_data_line(line, parser=float)
        for line in lines if line.strip() != ""
    ]
    file_data["data"] = dict(
        zip(header_items, zip(*data))
    )

    return file_data


def get_sanitize_data(file_data: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Data as found in the curve files are slightly different then
    the dictionary as expected by the 'upload_curve' method of the
    driver
    """
    data_dict = dict(file_data["data"])
    # We do not need the index column
    del data_dict["No."]
    # Rename the 'Units' column to the appropriate name
    # Look up under the 'Data Format' entry to find what units we have
    data_format = file_data['metadata']['Data Format']
    # This is a string in the form '4      (Log Ohms/Kelvin)'
    data_format_int = int(data_format.split()[0])
    correct_name = Model_325_Curve.valid_sensor_units[data_format_int - 1]
    # Rename the column
    data_dict[correct_name] = data_dict["Units"]
    del data_dict["Units"]

    return data_dict


class Status(IntFlag):
    sensor_units_overrang = 128
    sensor_units_zero = 64
    temp_overrange = 32
    temp_underrange = 16
    invalid_reading = 1


class Model_325_Curve(InstrumentChannel):

    valid_sensor_units = ["mV", "V", "Ohm", "log Ohm"]
    temperature_key = "Temperature (K)"

    def __init__(self, parent: 'Model_325', index: int) -> None:

        self._index = index
        name = f"curve_{index}"
        super().__init__(parent, name)

        self.add_parameter(
            "serial_number",
            parameter_class=GroupParameter
        )

        self.add_parameter(
            "format",
            val_mapping={
                f"{unt}/K": i+1 for i, unt in enumerate(self.valid_sensor_units)
            },
            parameter_class=GroupParameter
        )

        self.add_parameter(
            "limit_value",
            parameter_class=GroupParameter
        )

        self.add_parameter(
            "coefficient",
            val_mapping={
                "negative": 1,
                "positive": 2
            },
            parameter_class=GroupParameter
        )

        self.add_parameter(
            "curve_name",
            parameter_class=GroupParameter
        )

        Group(
            [
                self.curve_name, self.serial_number, self.format,
                self.limit_value, self.coefficient
            ],
            set_cmd=f"CRVHDR {self._index}, {{curve_name}}, "
                    f"{{serial_number}}, {{format}}, {{limit_value}}, "
                    f"{{coefficient}}",
            get_cmd=f"CRVHDR? {self._index}"
        )

    def get_data(self) -> Dict[Any, Any]:
        curve = [
            float(a) for point_index in range(1, 200)
            for a in self.ask(f"CRVPT? {self._index}, {point_index}").split(",")
        ]

        d = {self.temperature_key: curve[1::2]}
        sensor_unit = self.format().split("/")[0]
        d[sensor_unit] = curve[::2]

        return d

    @classmethod
    def validate_datadict(cls, data_dict: Dict[Any, Any]) -> str:
        """
        A data dict has two keys, one of which is 'Temperature (K)'. The other
        contains the units in which the curve is defined and must be one of:
        'mV', 'V', 'Ohm' or 'log Ohm'

        This method validates this and returns the sensor unit encountered in
        the data dict
        """
        if cls.temperature_key not in data_dict:
            raise ValueError(f"At least {cls.temperature_key} needed in the "
                             f"data dictionary")

        sensor_units = [i for i in data_dict.keys() if i != cls.temperature_key]

        if len(sensor_units) != 1:
            raise ValueError(
                "Data dictionary should have one other key, other then "
                "'Temperature (K)'"
            )

        sensor_unit = sensor_units[0]

        if sensor_unit not in cls.valid_sensor_units:
            raise ValueError(
                f"Sensor unit {sensor_unit} invalid. This needs to be one of "
                f"{', '.join(cls.valid_sensor_units)}"
            )

        data_size = len(data_dict[cls.temperature_key])
        if data_size != len(data_dict[sensor_unit]) or data_size > 200:
            raise ValueError("The length of the temperature axis should be "
                             "the same as the length of the sensor axis and "
                             "should not exceed 200 in size")

        return sensor_unit

    def set_data(self, data_dict: Dict[Any, Any],
                 sensor_unit: Optional[str] = None) -> None:
        """
        Set the curve data according to the values found the the dictionary.

        Args:
            data_dict (dict): See `validate_datadict` to see the format of this
                                dictionary
            sensor_unit (str): If None, the data dict is validated and the
                                units are extracted.
        """
        if sensor_unit is None:
            sensor_unit = self.validate_datadict(data_dict)

        temperature_values = data_dict[self.temperature_key]
        sensor_values = data_dict[sensor_unit]

        for value_index, (temperature_value, sensor_value) in \
                enumerate(zip(temperature_values, sensor_values)):

            cmd_str = f"CRVPT {self._index}, {value_index + 1}, " \
                      f"{sensor_value:3.3f}, {temperature_value:3.3f}"

            self.write(cmd_str)


class Model_325_Sensor(InstrumentChannel):
    """
    A single sensor of a  Lakeshore 325.

    Args:
        parent (Model_325): The instrument this heater belongs to
        name (str)
        inp (str): Either "A" or "B"
    """

    def __init__(self, parent: 'Model_325', name: str, inp: str) -> None:

        if inp not in ["A", "B"]:
            raise ValueError("Please either specify input 'A' or 'B'")

        super().__init__(parent, name)
        self._input = inp

        self.add_parameter(
            'temperature',
            get_cmd=f'KRDG? {self._input}',
            get_parser=float,
            label='Temperature',
            unit='K'
        )

        self.add_parameter(
            'status',
            get_cmd=f'RDGST? {self._input}',
            get_parser=lambda status: self.decode_sensor_status(int(status)),
            label='Sensor_Status'
        )

        self.add_parameter(
            "type",
            val_mapping={
                "Silicon diode": 0,
                "GaAlAs diode": 1,
                "100 Ohm platinum/250": 2,
                "100 Ohm platinum/500": 3,
                "1000 Ohm platinum": 4,
                "NTC RTD": 5,
                "Thermocouple 25mV": 6,
                "Thermocouple 50 mV": 7,
                "2.5 V, 1 mA": 8,
                "7.5 V, 1 mA": 9
            },
            parameter_class=GroupParameter
        )

        self.add_parameter(
            "compensation",
            vals=Enum(0, 1),
            parameter_class=GroupParameter
        )

        Group(
            [self.type, self.compensation],
            set_cmd=f"INTYPE {self._input}, {{type}}, {{compensation}}",
            get_cmd=f"INTYPE? {self._input}"
        )

        self.add_parameter(
            "curve_index",
            set_cmd=f"INCRV {self._input}, {{}}",
            get_cmd=f"INCRV? {self._input}",
            get_parser=int,
            vals=Numbers(min_value=1, max_value=35)
        )

    @staticmethod
    def decode_sensor_status(sum_of_codes: int) -> str:
        total_status = Status(sum_of_codes)
        if sum_of_codes == 0:
            return 'OK'
        status_messages = [st.name.replace('_', ' ') for st in Status
                           if st in total_status]
        return ", ".join(status_messages)

    @property
    def curve(self) -> Model_325_Curve:
        parent = cast(Model_325, self.parent)
        return Model_325_Curve(parent,  self.curve_index())


class Model_325_Heater(InstrumentChannel):
    """
    Heater control for the Lakeshore 325.

    Args:
        parent (Model_325): The instrument this heater belongs to
        name (str)
        loop (int): Either 1 or 2
    """
    def __init__(self, parent: 'Model_325', name: str, loop: int) -> None:

        if loop not in [1, 2]:
            raise ValueError("Please either specify loop 1 or 2")

        super().__init__(parent, name)
        self._loop = loop

        self.add_parameter(
            "control_mode",
            get_cmd=f"CMODE? {self._loop}",
            set_cmd=f"CMODE {self._loop},{{}}",
            val_mapping={
                "Manual PID": "1",
                "Zone": "2",
                "Open Loop": "3",
                "AutoTune PID": "4",
                "AutoTune PI": "5",
                "AutoTune P": "6"
            }
        )

        self.add_parameter(
            "input_channel",
            vals=Enum("A", "B"),
            parameter_class=GroupParameter
        )

        self.add_parameter(
            "unit",
            val_mapping={
                "Kelvin": "1",
                "Celsius": "2",
                "Sensor Units": "3"
            },
            parameter_class=GroupParameter
        )

        self.add_parameter(
            'powerup_enable',
            val_mapping={True: 1, False: 0},
            parameter_class=GroupParameter
        )

        self.add_parameter(
            "output_metric",
            val_mapping={
                "current": "1",
                "power": "2",
            },
            parameter_class=GroupParameter
        )

        Group(
            [self.input_channel, self.unit, self.powerup_enable,
             self.output_metric],
            set_cmd=f"CSET {self._loop}, {{input_channel}}, {{unit}}, "
                    f"{{powerup_enable}}, {{output_metric}}",
            get_cmd=f"CSET? {self._loop}"
        )

        self.add_parameter(
            'P',
            vals=Numbers(0, 1000),
            get_parser=float,
            parameter_class=GroupParameter
        )

        self.add_parameter(
            'I',
            vals=Numbers(0, 1000),
            get_parser=float,
            parameter_class=GroupParameter
        )

        self.add_parameter(
            'D',
            vals=Numbers(0, 1000),
            get_parser=float,
            parameter_class=GroupParameter
        )

        Group(
            [self.P, self.I, self.D],
            set_cmd=f'PID {self._loop}, {{P}}, {{I}}, {{D}}',
            get_cmd=f'PID? {self._loop}'
        )

        if self._loop == 1:
            valid_output_ranges = Enum(0, 1, 2)
        else:
            valid_output_ranges = Enum(0, 1)

        self.add_parameter(
            'output_range',
            vals=valid_output_ranges,
            set_cmd=f'RANGE {self._loop}, {{}}',
            get_cmd=f'RANGE? {self._loop}',
            val_mapping={
                "Off": '0',
                "Low (2.5W)": '1',
                "High (25W)": '2'
            }
        )

        self.add_parameter(
            'setpoint',
            vals=Numbers(0, 400),
            get_parser=float,
            set_cmd=f'SETP {self._loop}, {{}}',
            get_cmd=f'SETP? {self._loop}'
        )

        self.add_parameter(
            "ramp_state",
            vals=Enum(0, 1),
            parameter_class=GroupParameter
        )

        self.add_parameter(
            "ramp_rate",
            vals=Numbers(0, 100 / 60 * 1E3),
            unit="mK/s",
            parameter_class=GroupParameter,
            get_parser=lambda v: float(v) / 60 * 1E3,  # We get values in K/min,
            set_parser=lambda v: v * 60 * 1E-3  # Convert to K/min
        )

        Group(
            [self.ramp_state, self.ramp_rate],
            set_cmd=f"RAMP {self._loop}, {{ramp_state}}, {{ramp_rate}}",
            get_cmd=f"RAMP? {self._loop}"
        )

        self.add_parameter(
            "is_ramping",
            get_cmd=f"RAMPST? {self._loop}"
        )

        self.add_parameter(
            "resistance",
            get_cmd=f"HTRRES? {self._loop}",
            set_cmd=f"HTRRES {self._loop}, {{}}",
            val_mapping={
                25: 1,
                50: 2,
            },
            label='Resistance',
            unit="Ohm"
        )

        self.add_parameter(
            "heater_output",
            get_cmd=f"HTR? {self._loop}",
            get_parser=float,
            label='Heater Output',
            unit="%"
        )


class Model_325(VisaInstrument):
    """
    Lakeshore Model 325 Temperature Controller Driver
    """

    def __init__(self, name: str, address: str, **kwargs: Any) -> None:
        super().__init__(name, address, terminator="\r\n", **kwargs)

        sensors = ChannelList(
            self, "sensor", Model_325_Sensor, snapshotable=False)

        for inp in ['A', 'B']:
            sensor = Model_325_Sensor(self, f'sensor_{inp}', inp)
            sensors.append(sensor)
            self.add_submodule(f'sensor_{inp}', sensor)

        sensors.lock()
        self.add_submodule("sensor", sensors)

        heaters = ChannelList(
            self, "heater", Model_325_Heater, snapshotable=False)

        for loop in [1, 2]:
            heater = Model_325_Heater(self, f'heater_{loop}', loop)
            heaters.append(heater)
            self.add_submodule(f'heater_{loop}', heater)

        heaters.lock()
        self.add_submodule("heater", heaters)

        curves = ChannelList(
            self, "curve", Model_325_Curve, snapshotable=False
        )

        for curve_index in range(1, 35):
            curve = Model_325_Curve(self, curve_index)
            curves.append(curve)

        self.add_submodule("curve", curves)

        self.connect_message()

    def upload_curve(
            self, index: int, name: str,
            serial_number: str, data_dict: Dict[Any, Any]
    ) -> None:
        """
        Upload a curve to the given index

        Args:
             index: The index to upload the curve to. We can only use
                            indices reserved for user defined curves, 21-35
             name
             serial_number
             data_dict: A dictionary containing the curve data
        """
        if index not in range(21, 36):
            raise ValueError("index value should be between 21 and 35")

        sensor_unit = Model_325_Curve.validate_datadict(data_dict)

        curve = self.curve[index - 1]
        curve.curve_name(name)
        curve.serial_number(serial_number)
        curve.format(f"{sensor_unit}/K")
        curve.set_data(data_dict, sensor_unit=sensor_unit)

    def upload_curve_from_file(self, index: int, file_path: str) -> None:
        """
        Upload a curve from a curve file. Note that we only support
        curve files with extension .330
        """
        if not file_path.endswith(".330"):
            raise ValueError("Only curve files with extension .330 are supported")

        with open(file_path) as curve_file:
            file_data = read_curve_file(curve_file)

        data_dict = get_sanitize_data(file_data)
        name = file_data["metadata"]["Sensor Model"]
        serial_number = file_data["metadata"]["Serial Number"]

        self.upload_curve(index, name, serial_number, data_dict)



from typing import Any

from qcodes.utils import validators

from .ATS import AlazarTech_ATS
from .utils import TraceParameter


class AlazarTech_ATS9870(AlazarTech_ATS):
    """
    This class is the driver for the ATS9870 board
    it inherits from the ATS base class

    It creates all necessary parameters for the Alazar card
    """
    def __init__(self, name: str,
                 dll_path: str = 'C:\\WINDOWS\\System32\\ATSApi.dll',
                 **kwargs: Any):
        super().__init__(name, dll_path=dll_path, **kwargs)
        # add parameters

        # ----- Parameters for the configuration of the board -----
        self.add_parameter(name='clock_source',
                           parameter_class=TraceParameter,
                           label='Clock Source',
                           unit=None,
                           initial_value='INTERNAL_CLOCK',
                           val_mapping={'INTERNAL_CLOCK': 1,
                                        'SLOW_EXTERNAL_CLOCK': 4,
                                        'EXTERNAL_CLOCK_AC': 5,
                                        'EXTERNAL_CLOCK_10MHz_REF': 7})

        self.add_parameter(name='external_sample_rate',
                           get_cmd=None,
                           parameter_class=TraceParameter,
                           label='External Sample Rate',
                           unit='S/s',
                           vals=validators.Enum(1_000_000_000),
                           initial_value=1_000_000_000)

        self.add_parameter(name='sample_rate',
                           parameter_class=TraceParameter,
                           label='Sample Rate',
                           unit='S/s',
                           initial_value=1000000000,
                           val_mapping={
                               1000: 0x1,
                               2000: 0x2,
                               5000: 0x4,
                               10000: 0x8,
                               20000: 0xA,
                               50000: 0xC,
                               100000: 0xE,
                               200000: 0x10,
                               500000: 0x12,
                               1000000: 0x14,
                               2000000: 0x18,
                               5000000: 0x1A,
                               10000000: 0x1C,
                               20000000: 0x1E,
                               50000000: 0x22,
                               100000000: 0x24,
                               250000000: 0x2B,
                               500000000: 0x30,
                               1000000000: 0x35,
                               'EXTERNAL_CLOCK': 0x40,
                               '1GHz_REFERENCE_CLOCK': 1000000000})

        self.add_parameter(name='clock_edge',
                           parameter_class=TraceParameter,
                           label='Clock Edge',
                           unit=None,
                           initial_value='CLOCK_EDGE_RISING',
                           val_mapping={'CLOCK_EDGE_RISING': 0,
                                        'CLOCK_EDGE_FALLING': 1})

        self.add_parameter(name='decimation',
                           parameter_class=TraceParameter,
                           label='Decimation',
                           unit=None,
                           initial_value=0,
                           vals=validators.Ints(0, 100000))

        for i in range(1, self.channels + 1):
            self.add_parameter(name=f'coupling{i}',
                               parameter_class=TraceParameter,
                               label=f'Coupling channel {i}',
                               unit=None,
                               initial_value='AC',
                               val_mapping={'AC': 1, 'DC': 2})
            self.add_parameter(name=f'channel_range{i}',
                               parameter_class=TraceParameter,
                               label=f'Range channel {i}',
                               unit='V',
                               initial_value=4,
                               val_mapping={0.04: 2,
                                            0.1: 5,
                                            0.2: 6,
                                            0.4: 7,
                                            1.0: 10,
                                            2.0: 11,
                                            4.0: 12})
            self.add_parameter(name=f'impedance{i}',
                               parameter_class=TraceParameter,
                               label=f'Impedance channel {i}',
                               unit='Ohm',
                               initial_value=50,
                               val_mapping={1000000: 1,
                                            50: 2})
            self.add_parameter(name=f'bwlimit{i}',
                               parameter_class=TraceParameter,
                               label=f'Bandwidth limit channel {i}',
                               unit=None,
                               initial_value='DISABLED',
                               val_mapping={'DISABLED': 0,
                                            'ENABLED': 1})

        self.add_parameter(name='trigger_operation',
                           parameter_class=TraceParameter,
                           label='Trigger Operation',
                           unit=None,
                           initial_value='TRIG_ENGINE_OP_J',
                           val_mapping={'TRIG_ENGINE_OP_J': 0,
                                        'TRIG_ENGINE_OP_K': 1,
                                        'TRIG_ENGINE_OP_J_OR_K': 2,
                                        'TRIG_ENGINE_OP_J_AND_K': 3,
                                        'TRIG_ENGINE_OP_J_XOR_K': 4,
                                        'TRIG_ENGINE_OP_J_AND_NOT_K': 5,
                                        'TRIG_ENGINE_OP_NOT_J_AND_K': 6})
        n_trigger_engines = 2
        for i in range(1, n_trigger_engines+1):
            self.add_parameter(name=f'trigger_engine{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Engine {i}',
                               unit=None,
                               initial_value='TRIG_ENGINE_' + ('J' if i == 1 else 'K'),
                               val_mapping={'TRIG_ENGINE_J': 0,
                                            'TRIG_ENGINE_K': 1})
            self.add_parameter(name=f'trigger_source{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Source {i}',
                               unit=None,
                               initial_value='DISABLE',
                               val_mapping={'CHANNEL_A': 0,
                                            'CHANNEL_B': 1,
                                            'EXTERNAL': 2,
                                            'DISABLE': 3})
            self.add_parameter(name=f'trigger_slope{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Slope {i}',
                               unit=None,
                               initial_value='TRIG_SLOPE_POSITIVE',
                               val_mapping={'TRIG_SLOPE_POSITIVE': 1,
                                            'TRIG_SLOPE_NEGATIVE': 2})
            self.add_parameter(name=f'trigger_level{i}',
                               parameter_class=TraceParameter,
                               label=f'Trigger Level {i}',
                               unit=None,
                               initial_value=128,
                               vals=validators.Ints(0, 255))

        self.add_parameter(name='external_trigger_coupling',
                           parameter_class=TraceParameter,
                           label='External Trigger Coupling',
                           unit=None,
                           initial_value='DC',
                           val_mapping={'DC': 2})

        self.add_parameter(name='external_trigger_range',
                           parameter_class=TraceParameter,
                           label='External Trigger Range',
                           unit=None,
                           initial_value='ETR_5V',
                           val_mapping={'ETR_5V': 0,
                                        'ETR_1V': 1})
        self.add_parameter(name='trigger_delay',
                           parameter_class=TraceParameter,
                           label='Trigger Delay',
                           unit='Sample clock cycles',
                           initial_value=0,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='timeout_ticks',
                           parameter_class=TraceParameter,
                           label='Timeout Ticks',
                           unit='10 us',
                           initial_value=0,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='aux_io_mode',
                           parameter_class=TraceParameter,
                           label='AUX I/O Mode',
                           unit=None,
                           initial_value='AUX_IN_AUXILIARY',
                           val_mapping={'AUX_OUT_TRIGGER': 0,
                                        'AUX_IN_TRIGGER_ENABLE': 1,
                                        'AUX_IN_AUXILIARY': 13})

        self.add_parameter(name='aux_io_param',
                           parameter_class=TraceParameter,
                           label='AUX I/O Param',
                           unit=None,
                           initial_value='NONE',
                           val_mapping={'NONE': 0,
                                        'TRIG_SLOPE_POSITIVE': 1,
                                        'TRIG_SLOPE_NEGATIVE': 2})

        # ----- Parameters for the acquire function -----
        self.add_parameter(name='mode',
                           label='Acquisition mode',
                           unit=None,
                           initial_value='NPT',
                           set_cmd=None,
                           val_mapping={'NPT': 0x200, 'TS': 0x400})

        # samples_per_record must be a multiple of of some number (64 in the
        # case of ATS9870) and and has some minimum (256 in the case of ATS9870)
        # These values can be found in the ATS-SDK programmar's guide
        self.add_parameter(name='samples_per_record',
                           label='Samples per Record',
                           unit=None,
                           initial_value=96000,
                           set_cmd=None,
                           vals=validators.Multiples(
                                divisor=64, min_value=256))

        self.add_parameter(name='records_per_buffer',
                           label='Records per Buffer',
                           unit=None,
                           initial_value=1,
                           set_cmd=None,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='buffers_per_acquisition',
                           label='Buffers per Acquisition',
                           unit=None,
                           set_cmd=None,
                           initial_value=1,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='channel_selection',
                           label='Channel Selection',
                           unit=None,
                           set_cmd=None,
                           initial_value='AB',
                           val_mapping={'A': 1, 'B': 2, 'AB': 3})
        self.add_parameter(name='transfer_offset',
                           label='Transfer Offset',
                           unit='Samples',
                           set_cmd=None,
                           initial_value=0,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='external_startcapture',
                           label='External Startcapture',
                           unit=None,
                           set_cmd=None,
                           initial_value='ENABLED',
                           val_mapping={'DISABLED': 0X0,
                                        'ENABLED': 0x1})
        self.add_parameter(name='enable_record_headers',
                           label='Enable Record Headers',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x8})
        self.add_parameter(name='alloc_buffers',
                           label='Alloc Buffers',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x20})
        self.add_parameter(name='fifo_only_streaming',
                           label='Fifo Only Streaming',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x800})
        self.add_parameter(name='interleave_samples',
                           label='Interleave Samples',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x1000})
        self.add_parameter(name='get_processed_data',
                           label='Get Processed Data',
                           unit=None,
                           set_cmd=None,
                           initial_value='DISABLED',
                           val_mapping={'DISABLED': 0x0,
                                        'ENABLED': 0x2000})
        self.add_parameter(name='allocated_buffers',
                           label='Allocated Buffers',
                           unit=None,
                           set_cmd=None,
                           initial_value=1,
                           vals=validators.Ints(min_value=0))
        self.add_parameter(name='buffer_timeout',
                           label='Buffer Timeout',
                           unit='ms',
                           set_cmd=None,
                           initial_value=1000,
                           vals=validators.Ints(min_value=0))

        model = self.get_idn()['model']
        if model != 'ATS9870':
            raise Exception(f"The Alazar board kind is not 'ATS9870',"
                            f" found '{str(model)}' instead.")



import logging
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union, cast

import numpy as np
from typing_extensions import TypedDict

from qcodes.dataset.data_set import load_by_id
from qcodes.dataset.data_set_protocol import DataSetProtocol
from qcodes.dataset.descriptions.param_spec import ParamSpecBase
from qcodes.utils.deprecate import deprecate
from qcodes.utils.numpy_utils import list_of_data_to_maybe_ragged_nd_array

log = logging.getLogger(__name__)


class DSPlotData(TypedDict):
    """
    The dictionary used to represent data for use within `plot_dataset`
    """
    name: str
    unit: str
    label: str
    data: np.ndarray
    shape: Optional[Tuple[int, ...]]


@deprecate(alternative="ndarray.flatten()")
def flatten_1D_data_for_plot(rawdata: Union[Sequence[Sequence[Any]],
                                            np.ndarray]) -> np.ndarray:
    """
    Cast the return value of the database query to
    a 1D numpy array

    Args:
        rawdata: The return of the get_values function

    Returns:
        A one-dimensional numpy array

    """
    dataarray = np.array(rawdata).flatten()
    return dataarray


@deprecate(alternative="dataset.get_parameter_data")
def get_data_by_id(run_id: int) -> List[List[DSPlotData]]:
    """
    Load data from database and reshapes into 1D arrays with minimal
    name, unit and label metadata.
    Only returns data from parameters that depend on other parameters or
    parameters that other parameters depend on, i.e. data for standalone
    parameters are not returned.

    Args:
        run_id: run ID from the database

    Returns:
        a list of lists of dictionaries like this:

    ::

        [
          # each element in this list refers
          # to one dependent (aka measured) parameter
            [
              # each element in this list refers
              # to one independent (aka setpoint) parameter
              # that the dependent parameter depends on;
              # a dictionary with the data and metadata of the dependent
              # parameter is in the *last* element in this list
                ...
                {
                    'data': <1D numpy array of points>,
                    'name': <name of the parameter>,
                    'label': <label of the parameter or ''>,
                    'unit': <unit of the parameter or ''>
                },
                ...
            ],
            ...
        ]

    """
    ds = load_by_id(run_id)
    output = _get_data_from_ds(ds)
    return output


def _get_data_from_ds(ds: DataSetProtocol) -> List[List[DSPlotData]]:
    dependent_parameters: Tuple[ParamSpecBase, ...] = tuple(
        ds.description.interdeps.dependencies.keys()
    )

    all_data = ds.cache.data()

    parameter_data = {ps.name: all_data[ps.name] for ps in dependent_parameters}

    output = []

    for dep_name, data_dict in parameter_data.items():
        data_dicts_list = []

        dependent = ds.description.interdeps[dep_name]
        dependencies = ds.description.interdeps.dependencies[dependent]

        for param_spec_base in dependencies + (dependent,):
            my_data_dict: DSPlotData = {
                "name": param_spec_base.name,
                "unit": param_spec_base.unit,
                "label": param_spec_base.label,
                "data": data_dict[param_spec_base.name],
                "shape": None,
            }
            data_dicts_list.append(my_data_dict)

        if ds.description.shapes is not None:
            data_dicts_list[-1]["shape"] = ds.description.shapes.get(dependent.name)

        output.append(data_dicts_list)

    return output


def _all_steps_multiples_of_min_step(rows: np.ndarray) -> bool:
    """
    Are all steps integer multiples of the smallest step?
    This is used in determining whether the setpoints correspond
    to a regular grid

    Args:
        rows: the output of _rows_from_datapoints

    Returns:
        The answer to the question
    """

    steps_list: List[np.ndarray] = []
    for row in rows:
        # TODO: What is an appropriate precision?
        steps_list += list(np.unique(np.diff(row).round(decimals=15)))

    steps = np.unique(steps_list)
    remainders = np.mod(steps[1:]/steps[0], 1)

    # TODO: What are reasonable tolerances for allclose?
    asmoms = bool(np.allclose(remainders, np.zeros_like(remainders)))

    return asmoms


def _rows_from_datapoints(inputsetpoints: np.ndarray) -> np.ndarray:
    """
    Cast the (potentially) unordered setpoints into rows
    of sorted, unique setpoint values. Because of the way they are ordered,
    these rows do not necessarily correspond to actual rows of the scan,
    but they can nonetheless be used to identify certain scan types

    Args:
        setpoints: The raw setpoints as a one-dimensional array

    Returns:
        A ndarray of the rows
    """

    rows = []
    setpoints = inputsetpoints.copy()

    # first check if there is only one unique array in which case we can avoid the
    # potentially slow loop below
    temp, inds, count = np.unique(setpoints, return_index=True,
                                  return_counts=True)
    num_repeats_array = np.unique(count)
    if len(num_repeats_array) == 1 and count.sum() == len(inputsetpoints):
        return np.tile(temp, (num_repeats_array[0], 1))
    else:
        rows.append(temp)
        setpoints = np.delete(setpoints, inds)

    while len(setpoints) > 0:
        temp, inds = np.unique(setpoints, return_index=True)
        rows.append(temp)
        setpoints = np.delete(setpoints, inds)

    return list_of_data_to_maybe_ragged_nd_array(rows)


def _all_in_group_or_subgroup(rows: np.ndarray) -> bool:
    """
    Detects whether the setpoints correspond to two groups of
    of identical rows, one being contained in the other.

    This is the test for whether the setpoints correspond to a
    rectangular sweep. It allows for a single rectangular hole
    in the setpoint grid, thus allowing for an interrupted sweep.
    Note that each axis needs NOT be equidistantly spaced.

    Args:
        rows: The output from _rows_from_datapoints

    Returns:
        A boolean indicating whether the setpoints meet the
            criterion
    """

    groups = 1
    comp_to = rows[0]

    aigos = True
    switchindex = 0

    for rowind, row in enumerate(rows[1:]):
        if np.array_equal(row, comp_to):
            continue
        else:
            groups += 1
            comp_to = row
            switchindex = rowind
            if groups > 2:
                aigos = False
                break

    # if there are two groups, check that the rows of one group
    # are all contained in the rows of the other
    if aigos and switchindex > 0:
        for row in rows[1+switchindex:]:
            if sum(r in rows[0] for r in row) != len(row):
                aigos = False
                break

    return aigos


def _strings_as_ints(inputarray: np.ndarray) -> np.ndarray:
    """
    Return an integer-valued array version of a string-valued array. Maps, e.g.
    array(['a', 'b', 'c', 'a', 'c']) to array([0, 1, 2, 0, 2]). Useful for
    numerical setpoint analysis

    Args:
        inputarray: A 1D array of strings
    """
    newdata = np.zeros(len(inputarray))
    for n, word in enumerate(np.unique(inputarray)):
        newdata += ((inputarray == word).astype(int)*n)
    return newdata


def get_1D_plottype(xpoints: np.ndarray, ypoints: np.ndarray) -> str:
    """
    Determine plot type for a 1D plot by inspecting the data

    Possible plot types are:
    * '1D_bar' - bar plot
    * '1D_point' - scatter plot
    * '1D_line' - line plot

    Args:
        xpoints: The x-axis values
        ypoints: The y-axis values

    Returns:
        Determined plot type as a string
    """

    if isinstance(xpoints[0], str) and not isinstance(ypoints[0], str):
        if len(xpoints) == len(np.unique(xpoints)):
            return '1D_bar'
        else:
            return '1D_point'
    if isinstance(xpoints[0], str) or isinstance(ypoints[0], str):
        return '1D_point'
    else:
        return datatype_from_setpoints_1d(xpoints)


def datatype_from_setpoints_1d(setpoints: np.ndarray) -> str:
    """
    Figure out what type of visualisation is proper for the
    provided setpoints.

    The type is:
        * '1D_point' (scatter plot) when all setpoints are identical
        * '1D_line' otherwise

    Args:
        setpoints: The x-axis values

    Returns:
        A string representing the plot type as described above
    """
    if np.allclose(setpoints, setpoints[0]):
        return '1D_point'
    else:
        return '1D_line'


def get_2D_plottype(xpoints: np.ndarray,
                    ypoints: np.ndarray,
                    zpoints: np.ndarray) -> str:
    """
    Determine plot type for a 2D plot by inspecting the data

    Plot types are:
    * '2D_grid' - colormap plot for data that is on a grid
    * '2D_equidistant' - colormap plot for data that is on equidistant grid
    * '2D_scatter' - scatter plot
    * '2D_unknown' - returned in case the data did not match any criteria of the
    other plot types

    Args:
        xpoints: The x-axis values
        ypoints: The y-axis values
        zpoints: The z-axis (colorbar) values

    Returns:
        Determined plot type as a string
    """

    plottype = datatype_from_setpoints_2d(xpoints, ypoints)
    return plottype


def datatype_from_setpoints_2d(xpoints: np.ndarray,
                               ypoints: np.ndarray
                               ) -> str:
    """
    For a 2D plot, figure out what kind of visualisation we can use
    to display the data.

    Plot types are:
    * '2D_point' - all setpoint are the same in each direction; one point
    * '2D_grid' - colormap plot for data that is on a grid
    * '2D_equidistant' - colormap plot for data that is on equidistant grid
    * '2D_scatter' - scatter plot
    * '2D_unknown' - returned in case the data did not match any criteria of the
    other plot types

    Args:
        xpoints: The x-axis values
        ypoints: The y-axis values

    Returns:
        A string with the name of the determined plot type
    """
    # We represent categorical data as integer-valued data
    if isinstance(xpoints[0], str):
        xpoints = _strings_as_ints(xpoints)
    if isinstance(ypoints[0], str):
        ypoints = _strings_as_ints(ypoints)

    # First check whether all setpoints are identical along
    # any dimension
    x_all_the_same = np.allclose(xpoints, xpoints[0])
    y_all_the_same = np.allclose(ypoints, ypoints[0])

    if x_all_the_same or y_all_the_same:
        return '2D_point'

    # Now check if this is a simple rectangular sweep,
    # possibly interrupted in the middle of one row

    xrows = _rows_from_datapoints(xpoints)
    yrows = _rows_from_datapoints(ypoints)

    x_check = _all_in_group_or_subgroup(xrows)
    y_check = _all_in_group_or_subgroup(yrows)

    x_check = x_check and (len(xrows[0]) == len(yrows))
    y_check = y_check and (len(yrows[0]) == len(xrows))

    # this is the check that we are on a "simple" grid
    if y_check and x_check:
        return '2D_grid'

    x_check = _all_steps_multiples_of_min_step(xrows)
    y_check = _all_steps_multiples_of_min_step(yrows)

    # this is the check that we are on an equidistant grid
    if y_check and x_check:
        return '2D_equidistant'

    return '2D_unknown'


def reshape_2D_data(x: np.ndarray, y: np.ndarray, z: np.ndarray
                    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    xrow = np.array(_rows_from_datapoints(x)[0])
    yrow = np.array(_rows_from_datapoints(y)[0])
    nx = len(xrow)
    ny = len(yrow)

    # potentially slow method of filling in the data, should be optimised
    log.debug('Sorting 2D data onto grid')

    if isinstance(z[0], str):
        z_to_plot = np.full((ny, nx), '', dtype=z.dtype)
    else:
        z_to_plot = np.full((ny, nx), np.nan)
    x_index = np.zeros_like(x, dtype=np.dtype(np.int_))
    y_index = np.zeros_like(y, dtype=np.dtype(np.int_))
    for i, xval in enumerate(xrow):
        x_index[np.where(x == xval)[0]] = i
    for i, yval in enumerate(yrow):
        y_index[np.where(y == yval)[0]] = i

    z_to_plot[y_index, x_index] = z

    return xrow, yrow, z_to_plot


@deprecate(alternative="dataset.get_parameter_data")
def get_shaped_data_by_runid(
        run_id: int
) -> List[List[Dict[str, Union[str, np.ndarray]]]]:
    """
    Get data for a given run ID, but shaped according to its nature

    The data might get flattened, and additionally reshaped if it falls on a
    grid (equidistant or not).

    Args:
        run_id: The ID of the run for which to get data

    Returns:
        List of lists of dictionaries, the same as for `get_data_by_id`
    """
    mydata = get_data_by_id(run_id)

    for independet in mydata:
        data_length_long_enough = len(independet) == 3 \
                                  and len(independet[0]['data']) > 0 \
                                  and len(independet[1]['data']) > 0

        if data_length_long_enough:
            independet[0]['data'] = flatten_1D_data_for_plot(
                independet[0]['data'])
            independet[1]['data'] = flatten_1D_data_for_plot(
                independet[1]['data'])

            datatype = datatype_from_setpoints_2d(
                cast(np.ndarray, independet[0]['data']),
                cast(np.ndarray, independet[1]['data'])
            )

            if datatype in ('2D_grid', '2D_equidistant'):
                (
                    independet[0]['data'],
                    independet[1]['data'],
                    independet[2]['data']
                ) = reshape_2D_data(
                    cast(np.ndarray, independet[0]['data']),
                    cast(np.ndarray, independet[1]['data']),
                    cast(np.ndarray, independet[2]['data'])
                )

    return mydata



from typing import Any

from . import N52xx

class N5230C(N52xx.PNABase):
    def __init__(self, name: str, address: str, **kwargs: Any):
        super().__init__(name, address,
                         min_freq=300e3, max_freq=13.5e9,
                         min_power=-90, max_power=13,
                         nports=2,
                         **kwargs)



import logging
import time
from typing import Dict

from qcodes.instrument.base import InstrumentBase
from qcodes.instrument_drivers.Lakeshore.Model_336 import Model_336
import qcodes.instrument.sims as sims

from .test_lakeshore import MockVisaInstrument, split_args, DictClass, \
    query, command, instrument_fixture


log = logging.getLogger(__name__)

VISA_LOGGER = '.'.join((InstrumentBase.__module__, 'com', 'visa'))


class Model_336_Mock(MockVisaInstrument, Model_336):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # initial values
        self.heaters: Dict[str, DictClass] = {}
        self.heaters['1'] = DictClass(P=1, I=2, D=3,
                                      mode=1,  # 'off'
                                      input_channel=1,  # 'A'
                                      powerup_enable=0, polarity=0,
                                      use_filter=0, delay=1,
                                      output_range=0,
                                      setpoint=4)
        self.heaters['2'] = DictClass(P=1, I=2, D=3,
                                      mode=2,  # 'closed_loop'
                                      input_channel=2,  # 'B'
                                      powerup_enable=0, polarity=0,
                                      use_filter=0, delay=1,
                                      output_range=0,
                                      setpoint=4)
        self.heaters['3'] = DictClass(mode=4,  # 'monitor_out'
                                      input_channel=2,  # 'B'
                                      powerup_enable=0, polarity=0,
                                      use_filter=0, delay=1,
                                      output_range=0,
                                      setpoint=4)
        self.heaters['4'] = DictClass(mode=5,  # 'warm_up'
                                      input_channel=1,  # 'A'
                                      powerup_enable=0, polarity=0,
                                      use_filter=0, delay=1,
                                      output_range=0,
                                      setpoint=4)

        self.channel_mock = \
            {str(i): DictClass(t_limit=i, T=4,
                               sensor_name=f'sensor_{i}',
                               sensor_type=1,  # 'diode',
                               auto_range_enabled=0,  # 'off',
                               range=0,
                               compensation_enabled=0,  # False,
                               units=1)  # 'kelvin')
             for i in self.channel_name_command.keys()}

        # simulate delayed heating
        self.simulate_heating = False
        self.start_heating_time = time.perf_counter()

    def start_heating(self):
        self.start_heating_time = time.perf_counter()
        self.simulate_heating = True

    def get_t_when_heating(self):
        """
        Simply define a fixed setpoint of 4 k for now
        """
        delta = abs(time.perf_counter() - self.start_heating_time)
        # make it simple to start with: linear ramp 1K per second
        # start at 7K.
        return max(4, 7 - delta)

    @query('PID?')
    def pidq(self, arg):
        heater = self.heaters[arg]
        return f'{heater.P},{heater.I},{heater.D}'

    @command('PID')
    @split_args()
    def pid(self, output, P, I, D):
        for a, v in zip(['P', 'I', 'D'], [P, I, D]):
            setattr(self.heaters[output], a, v)

    @query('OUTMODE?')
    def outmodeq(self, arg):
        heater = self.heaters[arg]
        return (f'{heater.mode},{heater.input_channel},'
                f'{heater.powerup_enable}')

    @command('OUTMODE')
    @split_args()
    def outputmode(self, output, mode, input_channel, powerup_enable):
        h = self.heaters[output]
        h.output = output
        h.mode = mode
        h.input_channel = input_channel
        h.powerup_enable = powerup_enable

    @query('INTYPE?')
    def intypeq(self, channel):
        ch = self.channel_mock[channel]
        return (f'{ch.sensor_type},'
                f'{ch.auto_range_enabled},{ch.range},'
                f'{ch.compensation_enabled},{ch.units}')

    @command('INTYPE')
    @split_args()
    def intype(self, channel, sensor_type, auto_range_enabled,
               range_, compensation_enabled, units):
        ch = self.channel_mock[channel]
        ch.sensor_type = sensor_type
        ch.auto_range_enabled = auto_range_enabled
        ch.range = range_
        ch.compensation_enabled = compensation_enabled
        ch.units = units

    @query('RANGE?')
    def rangeq(self, heater):
        h = self.heaters[heater]
        return f'{h.output_range}'

    @command('RANGE')
    @split_args()
    def range_cmd(self, heater, output_range):
        h = self.heaters[heater]
        h.output_range = output_range

    @query('SETP?')
    def setpointq(self, heater):
        h = self.heaters[heater]
        return f'{h.setpoint}'

    @command('SETP')
    @split_args()
    def setpoint(self, heater, setpoint):
        h = self.heaters[heater]
        h.setpoint = setpoint

    @query('TLIMIT?')
    def tlimitq(self, channel):
        chan = self.channel_mock[channel]
        return f'{chan.tlimit}'

    @command('TLIMIT')
    @split_args()
    def tlimitcmd(self, channel, tlimit):
        chan = self.channel_mock[channel]
        chan.tlimit = tlimit

    @query('KRDG?')
    def temperature(self, output):
        chan = self.channel_mock[output]
        if self.simulate_heating:
            return self.get_t_when_heating()
        return f'{chan.T}'


@instrument_fixture(scope='function')
def lakeshore_336():
    visalib = sims.__file__.replace('__init__.py',
                                    'lakeshore_model336.yaml@sim')
    return Model_336_Mock('lakeshore_336_fixture', 'GPIB::2::INSTR',
                          visalib=visalib, device_clear=False)


def test_pid_set(lakeshore_336):
    ls = lakeshore_336
    P, I, D = 1, 2, 3
    # Only current source outputs/heaters have PID parameters,
    # voltages source outputs/heaters do not.
    outputs = [ls.output_1, ls.output_2]
    for h in outputs:  # a.k.a. heaters
        h.P(P)
        h.I(I)
        h.D(D)
        assert (h.P(), h.I(), h.D()) == (P, I, D)


def test_output_mode(lakeshore_336):
    ls = lakeshore_336
    mode = 'off'
    input_channel = 'A'
    powerup_enable = True
    outputs = [getattr(ls, f'output_{n}') for n in range(1, 5)]
    for h in outputs:  # a.k.a. heaters
        h.mode(mode)
        h.input_channel(input_channel)
        h.powerup_enable(powerup_enable)
        assert h.mode() == mode
        assert h.input_channel() == input_channel
        assert h.powerup_enable() == powerup_enable


def test_range(lakeshore_336):
    ls = lakeshore_336
    output_range = 'medium'
    outputs = [getattr(ls, f'output_{n}') for n in range(1, 5)]
    for h in outputs:  # a.k.a. heaters
        h.output_range(output_range)
        assert h.output_range() == output_range


def test_tlimit(lakeshore_336):
    ls = lakeshore_336
    tlimit = 5.1
    for ch in ls.channels:
        ch.t_limit(tlimit)
        assert ch.t_limit() == tlimit


def test_setpoint(lakeshore_336):
    ls = lakeshore_336
    setpoint = 5.1
    outputs = [getattr(ls, f'output_{n}') for n in range(1, 5)]
    for h in outputs:  # a.k.a. heaters
        h.setpoint(setpoint)
        assert h.setpoint() == setpoint


def test_select_range_limits(lakeshore_336):
    h = lakeshore_336.output_1
    ranges = [1, 2, 3]
    h.range_limits(ranges)

    for i in ranges:
        h.set_range_from_temperature(i - 0.5)
        assert h.output_range() == h.INVERSE_RANGES[i]

    h.set_range_from_temperature(i + 0.5)
    assert h.output_range() == h.INVERSE_RANGES[len(ranges)]


def test_set_and_wait_unit_setpoint_reached(lakeshore_336):
    ls = lakeshore_336
    ls.output_1.setpoint(4)
    ls.start_heating()
    ls.output_1.wait_until_set_point_reached()


def test_blocking_t(lakeshore_336):
    ls = lakeshore_336
    h = ls.output_1
    ranges = [1.2, 2.4, 3.1]
    h.range_limits(ranges)
    ls.start_heating()
    h.blocking_t(4)



"""
Live plotting in Jupyter notebooks
"""


class BasePlot:
    latest_plot = None
    """
    Auto-updating plot connected to a Jupyter notebook

    Args:
        interval (int): period in seconds between update checks
         default 1

        data_keys (str): sequence of keys in trace config can contain data
            that we should look for updates in.
            default 'xyz' (treated as a sequence) but add more if
            for example marker size or color can contain data
    """

    def __init__(self, interval=1, data_keys='xyz'):
        BasePlot.latest_plot = self
        self.data_keys = data_keys
        self.traces = []
        self.data_updaters = set()
        self.interval = interval
        self.standardunits = ['V', 's', 'J', 'W', 'm', 'eV', 'A', 'K', 'g',
                              'Hz', 'rad', 'T', 'H', 'F', 'Pa', 'C', 'Ω', 'Ohm',
                              'S']

    def clear(self):
        """
        Clears the plot window and removes all subplots and traces
        so that the window can be reused.
        """
        # any derived class should implement this
        raise NotImplementedError
        # typically traces and subplots should be cleared as well as the
        # figure window for the particular backend
        # TODO(giulioungaretti) the following unreachable lines should really
        # be documentation.
        self.traces = []
        self.subplots = []

    def replace(self, *args, updater=None, **kwargs):
        """
        Clear all content and add new trace.

        Args:
            args: optional way to provide x/y/z data without keywords
                If the last one is 1D, may be `y` or `x`, `y`
                If the last one is 2D, may be `z` or `x`, `y`, `z`

            updater: a callable (with no args) that updates the data in this trace
                if omitted, we will look for DataSets referenced in this data, and
                call their sync methods.

            **kwargs: passed on to self.add()
        """
        self.clear()
        self.add(*args, updater=updater, **kwargs)

    def add(self, *args, updater=None, **kwargs):
        """
        Add one trace to this plot.

        Args:
            args: optional way to provide x/y/z data without keywords
                If the last one is 1D, may be `y` or `x`, `y`
                If the last one is 2D, may be `z` or `x`, `y`, `z`

            updater: a callable (with no args) that updates the data in this trace
                if omitted, we will look for DataSets referenced in this data, and
                call their sync methods.

            kwargs: after inserting info found in args and possibly in set_arrays
                into `x`, `y`, and optionally `z`, these are passed along to
                self.add_to_plot.

        Returns:
            Plot handle for trace

        Examples:
            To use custom labels and units pass for example:

            >>> plot.add(x=set, y=amplitude,
            >>>          xlabel="set",
            >>>          xunit="V",
            >>>          ylabel= "Amplitude",
            >>>          yunit ="V")

        Array shapes for 2D plots:
            x:(1D-length m), y:(1D-length n), z: (2D- n*m array)
        """
        # TODO(giulioungaretti): replace with an explicit version, see expand trace
        self.expand_trace(args, kwargs)
        plot_object = self.add_to_plot(**kwargs)
        self.add_updater(updater, kwargs)

        return plot_object

    def add_to_plot(self, **kwargs):
        """
        Add a trace the plot itself (typically called by self.add,
        which incorporates args into kwargs, so the subclass doesn't
        need to worry about this). Data will be in `x`, `y`, and optionally
        `z`.

        Should be implemented by a subclass, and each call should append
        a dictionary to self.traces, containing at least {'config': kwargs}
        """
        raise NotImplementedError

    def add_updater(self, updater, plot_config):
        """
        Add an updater to the plot.

        Args:
            updater (Callable): callable (with no args) that updates the data in this trace
                if omitted, we will look for DataSets referenced in this data, and
                call their sync methods.
            plot_config (dict): this is a dictionary that gets populated inside
                add() via expand_trace().
                The reason this is here is to fetch from the data_set the sync method
                to use it as an updater.
        """
        if updater is not None:
            self.data_updaters.add(updater)
        else:
            for key in self.data_keys:
                data_array = plot_config.get(key, '')
                if hasattr(data_array, 'data_set'):
                    if data_array.data_set is not None:
                        self.data_updaters.add(data_array.data_set.sync)

        # If previous data on this plot became static, perhaps because
        # its measurement loop finished, the updater may have been halted.
        # If we have new update functions, re-activate the updater
        # by reinstating its update interval
        if self.data_updaters:
            if hasattr(self, 'update_widget'):
                self.update_widget.interval = self.interval

    def get_default_title(self):
        """
        Get the default title, which for a plot is just a list of DataSet locations.
        A custom title can be set when adding any trace (via either __init__ or add.
        these kwargs all eventually end up in self.traces[i]['config']) and it looks
        like we will take the first title we find from any trace... otherwise, if no
        trace specifies a title, then we combine whatever dataset locations we find.

        Note: (alexj): yeah, that's awkward, isn't it, and it looks like a weird
        implementation, feel free to change it 👼

        Returns:
            str: the title of the figure
        """
        title_parts = []
        for trace in self.traces:
            config = trace['config']
            if 'title' in config:  # can be passed using **kw
                return config['title']
            for part in self.data_keys:
                data_array = config.get(part, '')
                if hasattr(data_array, 'data_set'):
                    if data_array.data_set is not None:
                        location = data_array.data_set.location
                        if location and location not in title_parts:
                            title_parts.append(location)
        return ', '.join(title_parts)

    @staticmethod
    def get_label(data_array):
        """
        Look for a label in data_array falling back on name.

        Args:
            data_array (DataArray): data array to get label from

        Returns:
            str: label or name of the data_array

        """
        # TODO this should really be a static method
        name = (getattr(data_array, 'label', '') or
                getattr(data_array, 'name', ''))
        unit = getattr(data_array, 'unit', '')
        return  name, unit

    @staticmethod
    def expand_trace(args, kwargs):
        """
        Complete the x, y (and possibly z) data definition for a trace.

        Also modifies kwargs in place so that all the data needed to fully
        specify the trace is present (ie either x and y or x and y and z)

        Both ``__init__`` (for the first trace) and the ``add`` method support
        multiple ways to specify the data in the trace:

        As ``*args``:
            - ``add(y)`` or ``add(z)`` specify just the main 1D or 2D data, with
              the setpoint axis or axes implied.
            - ``add(x, y)`` or ``add(x, y, z)`` specify all axes of the data.
        And as ``**kwargs``:
            - ``add(x=x, y=y, z=z)`` you specify exactly the data you want on
              each axis. Any but the last (y or z) can be omitted, which allows
              for all of the same forms as with ``*args``, plus x and z or y and
              z, with just one axis implied from the setpoints of the z data.

        This method takes any of those forms and converts them into a complete
        set of kwargs, containing all of the explicit or implied data to be used
        in plotting this trace.

        Args:
            args (Tuple[DataArray]): positional args, as passed to either
                ``__init__`` or ``add``
            kwargs (Dict(DataArray]): keyword args, as passed to either
                ``__init__`` or ``add``. kwargs may contain non-data items in
                keys other than x, y, and z.

        Raises:
           ValueError: if the shape of the data does not match that of args
           ValueError: if the data is provided twice
        """
        # TODO(giulioungaretti): replace with an explicit version:
        # return the new kwargs  instead of modifying in place
        # TODO this should really be a static method
        if args:
            if hasattr(args[-1][0], '__len__'):
                # 2D (or higher... but ignore this for now)
                # this test works for both numpy arrays and bare sequences
                axletters = 'xyz'
                ndim = 2
            else:
                axletters = 'xy'
                ndim = 1

            if len(args) not in (1, len(axletters)):
                raise ValueError('{}D data needs 1 or {} unnamed args'.format(
                    ndim, len(axletters)))

            arg_axletters = axletters[-len(args):]

            for arg, arg_axletters in zip(args, arg_axletters):
                if arg_axletters in kwargs:
                    raise ValueError(arg_axletters + ' data provided twice')
                kwargs[arg_axletters] = arg

        # reset axletters, we may or may not have found them above
        axletters = 'xyz' if 'z' in kwargs else 'xy'
        main_data = kwargs[axletters[-1]]
        if hasattr(main_data, 'set_arrays'):
            num_axes = len(axletters) - 1
            # things will probably fail if we try to plot arrays of the
            # wrong dimension... but we'll give it a shot anyway.
            set_arrays = main_data.set_arrays[-num_axes:]
            # for 2D: y is outer loop, which is earlier in set_arrays,
            # and x is the inner loop... is this the right convention?
            set_axletters = reversed(axletters[:-1])
            for axletter, set_array in zip(set_axletters, set_arrays):
                if axletter not in kwargs:
                    kwargs[axletter] = set_array

    def update(self):
        """
        Update the data in this plot, using the updaters given with
        MatPlot.add() or in the included DataSets, then include this in
        the plot.

        This is a wrapper routine that the update widget calls,
        inside this we call self.update() which should be subclassed
        """
        any_updates = False
        for updater in self.data_updaters:
            updates = updater()
            if updates is not False:
                any_updates = True

        self.update_plot()

        # once all updaters report they're finished (by returning exactly
        # False) we stop updating the plot.
        if any_updates is False:
            self.halt()

    def update_plot(self):
        """
        Update the plot itself (typically called by self.update).
        Should be implemented by a subclass
        """
        raise NotImplementedError

    def halt(self):
        """
        Stop automatic updates to this plot, by canceling its update widget
        """
        if hasattr(self, 'update_widget'):
            self.update_widget.halt()

    def save(self, filename=None):
        """
        Save current plot to filename

        Args:
            filename (Optional[str]): Location of the file
        """
        raise NotImplementedError



"""
This module provides a number of convenient general-purpose functions that
are useful for building more database-specific queries out of them.
"""
import itertools
import sqlite3
from distutils.version import LooseVersion
from typing import Any, List, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
from numpy import ndarray

from qcodes.dataset.sqlite.connection import (
    ConnectionPlus,
    atomic,
    atomic_transaction,
    transaction,
)
from qcodes.dataset.sqlite.settings import SQLiteSettings
from qcodes.utils.deprecate import deprecate

# represent the type of  data we can/want map to sqlite column
VALUE = Union[str, complex, List, ndarray, bool, None]
VALUES = List[VALUE]


def one(curr: sqlite3.Cursor, column: Union[int, str]) -> Any:
    """Get the value of one column from one row
    Args:
        curr: cursor to operate on
        column: name of the column

    Returns:
        the value
    """
    res = curr.fetchall()
    if len(res) > 1:
        raise RuntimeError("Expected only one row")
    elif len(res) == 0:
        raise RuntimeError("Expected one row")
    else:
        return res[0][column]


def many(curr: sqlite3.Cursor, *columns: str) -> List[Any]:
    """Get the values of many columns from one row
    Args:
        curr: cursor to operate on
        columns: names of the columns

    Returns:
        list of  values
    """
    res = curr.fetchall()
    if len(res) > 1:
        raise RuntimeError("Expected only one row")
    else:
        return [res[0][c] for c in columns]


def many_many(curr: sqlite3.Cursor, *columns: str) -> List[List[Any]]:
    """Get all values of many columns
    Args:
        curr: cursor to operate on
        columns: names of the columns

    Returns:
        list of lists of values
    """
    res = curr.fetchall()
    results = []
    for r in res:
        results.append([r[c] for c in columns])
    return results


def select_one_where(
    conn: ConnectionPlus, table: str, column: str, where_column: str, where_value: VALUE
) -> VALUE:
    """
    Select a value from a given column given a match of a value in a
    different column. If the given matched row/column intersect is empty
    None will be returned.

    Args:
        conn: Connection to the db
        table: Table to look for values in
        column: Column to return value from
        where_column: Column to match on
        where_value: Value to match in where_column

    Returns:
        Value found
    raises:
        RuntimeError if not exactly match is found.
    """
    query = f"""
    SELECT {column}
    FROM
        {table}
    WHERE
        {where_column} = ?
    """
    cur = atomic_transaction(conn, query, where_value)
    res = one(cur, column)
    return res


def select_many_where(
    conn: ConnectionPlus,
    table: str,
    *columns: str,
    where_column: str,
    where_value: VALUE,
) -> VALUES:
    _columns = ",".join(columns)
    query = f"""
    SELECT {_columns}
    FROM
        {table}
    WHERE
        {where_column} = ?
    """
    cur = atomic_transaction(conn, query, where_value)
    res = many(cur, *columns)
    return res


def _massage_dict(metadata: Mapping[str, Any]) -> Tuple[str, List[Any]]:
    """
    {key:value, key2:value} -> ["key=?, key2=?", [value, value]]
    """
    template = []
    values = []
    for key, value in metadata.items():
        template.append(f"{key} = ?")
        values.append(value)
    return ','.join(template), values


def update_where(conn: ConnectionPlus, table: str,
                 where_column: str, where_value: Any, **updates: Any) -> None:
    _updates, values = _massage_dict(updates)
    query = f"""
    UPDATE
        '{table}'
    SET
        {_updates}
    WHERE
        {where_column} = ?
    """
    atomic_transaction(conn, query, *values, where_value)


def insert_values(conn: ConnectionPlus,
                  formatted_name: str,
                  columns: List[str],
                  values: VALUES,
                  ) -> int:
    """
    Inserts values for the specified columns.
    Will pad with null if not all parameters are specified.
    NOTE this need to be committed before closing the connection.
    """
    _columns = ",".join(columns)
    _values = ",".join(["?"] * len(columns))
    query = f"""INSERT INTO "{formatted_name}"
        ({_columns})
    VALUES
        ({_values})
    """

    c = atomic_transaction(conn, query, *values)
    return c.lastrowid


def insert_many_values(conn: ConnectionPlus,
                       formatted_name: str,
                       columns: Sequence[str],
                       values: Sequence[VALUES],
                       ) -> int:
    """
    Inserts many values for the specified columns.

    Example input:
    columns: ['xparam', 'yparam']
    values: [[x1, y1], [x2, y2], [x3, y3]]

    NOTE this need to be committed before closing the connection.
    """
    # We demand that all values have the same length
    lengths = [len(val) for val in values]
    if len(np.unique(lengths)) > 1:
        raise ValueError('Wrong input format for values. Must specify the '
                         'same number of values for all columns. Received'
                         f' lengths {lengths}.')
    no_of_rows = len(lengths)
    no_of_columns = lengths[0]

    # The TOTAL number of inserted values in one query
    # must be less than the SQLITE_MAX_VARIABLE_NUMBER

    # Version check cf.
    # "https://stackoverflow.com/questions/9527851/sqlite-error-
    #  too-many-terms-in-compound-select"
    version = SQLiteSettings.settings['VERSION']

    # According to the SQLite changelog, the version number
    # to check against below
    # ought to be 3.7.11, but that fails on Travis
    if LooseVersion(str(version)) <= LooseVersion('3.8.2'):
        max_var = SQLiteSettings.limits['MAX_COMPOUND_SELECT']
    else:
        max_var = SQLiteSettings.limits['MAX_VARIABLE_NUMBER']
    rows_per_transaction = int(int(max_var)/no_of_columns)

    _columns = ",".join(columns)
    _values = "(" + ",".join(["?"] * len(values[0])) + ")"

    a, b = divmod(no_of_rows, rows_per_transaction)
    chunks = a*[rows_per_transaction] + [b]
    if chunks[-1] == 0:
        chunks.pop()

    start = 0
    stop = 0

    with atomic(conn) as conn:
        for ii, chunk in enumerate(chunks):
            _values_x_params = ",".join([_values] * chunk)

            query = f"""INSERT INTO "{formatted_name}"
                        ({_columns})
                        VALUES
                        {_values_x_params}
                     """
            stop += chunk
            # we need to make values a flat list from a list of list
            flattened_values = list(
                itertools.chain.from_iterable(values[start:stop]))

            c = transaction(conn, query, *flattened_values)

            if ii == 0:
                return_value = c.lastrowid
            start += chunk

    return return_value


@deprecate('Unused private method to be removed in a future version')
def modify_values(conn: ConnectionPlus,
                  formatted_name: str,
                  index: int,
                  columns: List[str],
                  values: VALUES,
                  ) -> int:
    """
    Modify values for the specified columns.
    If a column is in the table but not in the columns list is
    left untouched.
    If a column is mapped to None, it will be a null value.
    """
    name_val_template = []
    for name in columns:
        name_val_template.append(f"{name}=?")
    name_val_templates = ",".join(name_val_template)
    query = f"""
    UPDATE "{formatted_name}"
    SET
        {name_val_templates}
    WHERE
        rowid = {index+1}
    """
    c = atomic_transaction(conn, query, *values)
    return c.rowcount


@deprecate('Unused private method to be removed in a future version')
def modify_many_values(conn: ConnectionPlus,
                       formatted_name: str,
                       start_index: int,
                       columns: List[str],
                       list_of_values: List[VALUES],
                       ) -> None:
    """
    Modify many values for the specified columns.
    If a column is in the table but not in the column list is
    left untouched.
    If a column is mapped to None, it will be a null value.
    """
    _len = length(conn, formatted_name)
    len_requested = start_index + len(list_of_values[0])
    available = _len - start_index
    if len_requested > _len:
        reason = f"""Modify operation Out of bounds.
        Trying to modify {len(list_of_values)} results,
        but therere are only {available} results.
        """
        raise ValueError(reason)
    for values in list_of_values:
        modify_values(conn, formatted_name, start_index, columns, values)
        start_index += 1


def length(conn: ConnectionPlus,
           formatted_name: str
           ) -> int:
    """
    Return the length of the table

    Args:
        conn: the connection to the sqlite database
        formatted_name: name of the table

    Returns:
        the length of the table
    """
    query = f"select MAX(id) from '{formatted_name}'"
    c = atomic_transaction(conn, query)
    _len = c.fetchall()[0][0]
    if _len is None:
        return 0
    else:
        return _len


def insert_column(conn: ConnectionPlus, table: str, name: str,
                  paramtype: Optional[str] = None) -> None:
    """Insert new column to a table

    Args:
        conn: database connection
        table: destination for the insertion
        name: column name
        type: sqlite type of the column
    """
    # first check that the column is not already there
    # and do nothing if it is
    query = f'PRAGMA TABLE_INFO("{table}");'
    cur = atomic_transaction(conn, query)
    columns = many_many(cur, "name")
    if name in [col[0] for col in columns]:
        return

    with atomic(conn) as conn:
        if paramtype:
            transaction(conn,
                        f'ALTER TABLE "{table}" ADD COLUMN "{name}" '
                        f'{paramtype}')
        else:
            transaction(conn,
                        f'ALTER TABLE "{table}" ADD COLUMN "{name}"')


def is_column_in_table(conn: ConnectionPlus, table: str, column: str) -> bool:
    """
    A look-before-you-leap function to look up if a table has a certain column.

    Intended for the 'runs' table where columns might be dynamically added
    via `add_meta_data`/`insert_meta_data` functions.

    Args:
        conn: The connection
        table: the table name
        column: the column name
    """
    cur = atomic_transaction(conn, f"PRAGMA table_info({table})")
    for row in cur.fetchall():
        if row['name'] == column:
            return True
    return False


def sql_placeholder_string(n: int) -> str:
    """
    Return an SQL value placeholder string for n values.
    Example: sql_placeholder_string(5) returns '(?,?,?,?,?)'
    """
    return '(' + ','.join('?'*n) + ')'



"""Tests for Alazar DLL API

This suite of tests is expected to be executed on a Windows PC with a single
Alazar board installed.
"""

import os
import gc
import logging

import pytest

from qcodes.instrument_drivers.AlazarTech.ATS import AlazarTech_ATS
from qcodes.instrument_drivers.AlazarTech.ats_api import AlazarATSAPI
from qcodes.instrument_drivers.AlazarTech.dll_wrapper import DllWrapperMeta
from qcodes.instrument_drivers.AlazarTech.constants import ERROR_CODES, \
    API_SUCCESS, Capability


def _skip_if_alazar_dll_and_boards_not_installed():
    if not os.path.exists(AlazarTech_ATS.dll_path + '.dll'):
        return pytest.mark.skip(
            "Alazar API DLL was not found in 'AlazarTech_ATS.dll_path'.")

    return pytest.mark.skipif(
        len(AlazarTech_ATS.find_boards()) != 1,
        reason='No, or more than one Alazar boards are installed on this PC.')


pytestmark = _skip_if_alazar_dll_and_boards_not_installed()


# Set the following constants to correct values, they are used in tests below.
SYSTEM_ID = 1
BOARD_ID = 1


@pytest.fixture
def alazar():
    alazar = AlazarTech_ATS('alazar', system_id=SYSTEM_ID, board_id=BOARD_ID)
    yield alazar
    alazar.close()


@pytest.fixture
def alazar_api():
    yield AlazarATSAPI(AlazarTech_ATS.dll_path)


def test_alazar_api_singleton_behavior(caplog):
    using_msg = lambda dll_path: (
        f"Using existing instance for DLL path {dll_path}.")
    creating_msg = lambda dll_path: (
        f"Creating new instance for DLL path {dll_path}.")

    assert DllWrapperMeta._instances == {}

    with caplog.at_level(logging.DEBUG):
        api1 = AlazarATSAPI(AlazarTech_ATS.dll_path)
    assert DllWrapperMeta._instances == {AlazarTech_ATS.dll_path: api1}
    assert caplog.records[-1].message == creating_msg(AlazarTech_ATS.dll_path)
    caplog.clear()

    with caplog.at_level(logging.DEBUG):
        api2 = AlazarATSAPI(AlazarTech_ATS.dll_path)
    assert api2 is api1
    assert DllWrapperMeta._instances == {AlazarTech_ATS.dll_path: api1}
    assert caplog.records[-1].message == using_msg(AlazarTech_ATS.dll_path)
    caplog.clear()

    # Indeed, this actually exposes a vulnarability of the setup. As far as
    # LoadLibrary from ctypes is concerned, both "..\AlazarApi" and
    # "..\AlazarApi.dll" would result in the same loaded library with even
    # the same `_handle` value. But here we will abuse this in order to create
    # a new instance of the Alazar API class by using the same DLL file.
    # This should probably be fixed.
    dll_path_3 = AlazarTech_ATS.dll_path + '.dll'
    with caplog.at_level(logging.DEBUG):
        api3 = AlazarATSAPI(dll_path_3)
    assert api3 is not api1
    assert api3 is not api2
    assert DllWrapperMeta._instances == {AlazarTech_ATS.dll_path: api1,
                                         dll_path_3: api3}
    assert caplog.records[-1].message == creating_msg(dll_path_3)
    caplog.clear()

    del api2
    gc.collect()
    assert DllWrapperMeta._instances == {AlazarTech_ATS.dll_path: api1,
                                         dll_path_3: api3}

    del api1
    gc.collect()
    assert DllWrapperMeta._instances == {dll_path_3: api3}

    del api3
    gc.collect()
    assert DllWrapperMeta._instances == {}


def test_find_boards():
    boards = AlazarTech_ATS.find_boards()
    assert len(boards) == 1
    assert boards[0]['system_id'] == SYSTEM_ID
    assert boards[0]['board_id'] == BOARD_ID


def test_get_board_info(alazar_api):
    info = AlazarTech_ATS.get_board_info(api=alazar_api,
                                         system_id=SYSTEM_ID,
                                         board_id=BOARD_ID)
    assert {'system_id', 'board_id', 'board_kind',
            'max_samples', 'bits_per_sample'} == set(list(info.keys()))
    assert info['system_id'] == SYSTEM_ID
    assert info['board_id'] == BOARD_ID


def test_idn(alazar):
    idn = alazar.get_idn()
    assert {'firmware', 'model', 'serial', 'vendor', 'CPLD_version',
            'driver_version', 'SDK_version', 'latest_cal_date', 'memory_size',
            'asopc_type', 'pcie_link_speed', 'pcie_link_width',
            'bits_per_sample', 'max_samples'
            } == set(list(idn.keys()))
    assert idn['vendor'] == 'AlazarTech'
    assert idn['model'][:3] == 'ATS'


def test_return_codes_are_correct(alazar_api):
    """
    Test correctness of the coded return codes (success, failure, unknowns),
    and consistency with what `AlazarErrorToText` function returns.
    """
    for code, msg in ERROR_CODES.items():
        real_msg = alazar_api.error_to_text(code)
        assert real_msg in msg

    assert alazar_api.error_to_text(API_SUCCESS) == 'ApiSuccess'

    lower_unknown = API_SUCCESS - 1
    assert alazar_api.error_to_text(lower_unknown) == 'Unknown'

    upper_unknown = max(list(ERROR_CODES.keys())) + 1
    assert alazar_api.error_to_text(upper_unknown) == 'Unknown'


def test_get_channel_info_convenient(alazar):
    bps, max_s = alazar.api.get_channel_info_(alazar._handle)
    assert isinstance(bps, int)
    assert isinstance(max_s, int)


def test_get_cpld_version_convenient(alazar):
    cpld_ver = alazar.api.get_cpld_version_(alazar._handle)
    assert isinstance(cpld_ver, str)
    assert len(cpld_ver.split('.')) == 2


def test_get_driver_version_convenient(alazar_api):
    driver_ver = alazar_api.get_driver_version_()
    assert isinstance(driver_ver, str)
    assert len(driver_ver.split('.')) == 3


def test_get_sdk_version_convenient(alazar_api):
    sdk_ver = alazar_api.get_sdk_version_()
    assert isinstance(sdk_ver, str)
    assert len(sdk_ver.split('.')) == 3


def test_query_capability_convenient(alazar):
    cap = Capability.GET_SERIAL_NUMBER
    cap_value = alazar.api.query_capability_(alazar._handle, cap)
    assert isinstance(cap_value, int)


def test_writing_and_reading_registers(alazar):
    """
    The approach is to read the register that includes information about
    trigger holdoff parameter, and write the same value back to the board.
    """
    trigger_holdoff_register_offset = 58
    orig_val = alazar._read_register(trigger_holdoff_register_offset)
    alazar._write_register(trigger_holdoff_register_offset, orig_val)


def test_get_num_channels():
    assert 1 == AlazarTech_ATS.get_num_channels(1)
    assert 1 == AlazarTech_ATS.get_num_channels(8)

    assert 2 == AlazarTech_ATS.get_num_channels(3)
    assert 2 == AlazarTech_ATS.get_num_channels(10)

    assert 4 == AlazarTech_ATS.get_num_channels(15)
    assert 8 == AlazarTech_ATS.get_num_channels(255)
    assert 16 == AlazarTech_ATS.get_num_channels(65535)

    with pytest.raises(RuntimeError, match='0'):
        AlazarTech_ATS.get_num_channels(0)
    with pytest.raises(RuntimeError, match='17'):
        AlazarTech_ATS.get_num_channels(17)
    with pytest.raises(RuntimeError, match='100'):
        AlazarTech_ATS.get_num_channels(100)
